import pytest

from nengo.exceptions import SpaParseError
from nengo.spa.action_objects import DotProduct, Source, Symbol

A = Symbol("A")
B = Symbol("B")
x = Source("x")
y = Source("y")


def test_symbol():
    A = Symbol("A")
    assert A.symbol == "A"
    B = Symbol("B")
    assert str(B) == "B"

    assert str(A + B) == "(A + B)"
    with pytest.raises(TypeError):
        A + "A"

    assert str(A - B) == "(A - B)"
    with pytest.raises(TypeError):
        A - []

    assert str(A * B) == "(A * B)"
    assert str(A * 1) == "A"
    assert str(1 * A) == "A"
    assert str(1.0 * A) == "A"
    assert str(A * 1.0) == "A"
    assert str(A * 1.5) == "(A * 1.5)"
    assert str(A * 2) == "(A * 2)"
    assert str(A * Symbol("1")) == "A"
    assert str(Symbol("1") * A) == "A"

    with pytest.raises(TypeError):
        A * {}

    assert str(~A) == "~A"
    assert str(~(~A)) == "A"

    assert str(-A) == "-A"
    assert str(-(-A)) == "A"

    assert str(A * B + (B * ~A) * 0.5 - A) == "(((A * B) + ((B * ~A) * 0.5)) - A)"


def test_source():
    A = Symbol("A")
    s = Source("s")
    assert s.name == "s"
    assert str(s.transform) == "1"
    assert str(s) == "s"
    s_q = Source("s", Symbol("Q"))
    assert s_q.name == "s"
    assert str(s_q.transform) == "Q"
    assert str(s_q) == "Q * s"

    assert str(s * 1.5) == "1.5 * s"
    assert str(s * A) == "A * s"
    assert str(s_q * A) == "(Q * A) * s"

    assert str(1.5 * s) == "1.5 * s"
    assert str(A * s) == "A * s"
    assert str(A * s_q) == "(Q * A) * s"

    assert str(-s) == "-1 * s"
    assert str(-(-s)) == "s"
    assert str(-s_q) == "-Q * s"

    assert str(s + s_q) == "s + Q * s"

    with pytest.raises(TypeError):
        s * "A"


def test_inverted():
    assert str(~x) == "~x"
    assert str(A * ~x) == "A * ~x"
    assert str(~x * y) == "((~x) * (y)) * 1"
    assert str(~x * ~y) == "((~x) * (~y)) * 1"
    assert str(x * ~y) == "((x) * (~y)) * 1"
    with pytest.raises(ValueError):
        ~(2 * x)
    with pytest.raises(ValueError):
        ~(A * x)


def test_dotproduct():
    assert str(DotProduct(x, y)) == "dot(x, y)"
    assert str(DotProduct(x, A)) == "dot(x, A)"
    assert str(DotProduct(B, y)) == "dot(B, y)"

    with pytest.raises(SpaParseError):
        DotProduct(B, A)
    with pytest.raises(SpaParseError):
        DotProduct(A, "B")
    with pytest.raises(SpaParseError):
        DotProduct(1, A)

    assert str(0.5 * DotProduct(x, y)) == "0.5 * dot(x, y)"
    assert str(0.5 * (0.5 * DotProduct(x, y))) == "0.25 * dot(x, y)"
    assert str(DotProduct(B, y) * 2) == "2 * dot(B, y)"

    with pytest.raises(TypeError):
        A * DotProduct(x, y)
    with pytest.raises(TypeError):
        DotProduct(x, y) * B

    assert str(DotProduct(x, y) / 2) == "0.5 * dot(x, y)"
    assert str((DotProduct(x, y) / 2) / 2) == "0.25 * dot(x, y)"
    assert str(2 * DotProduct(B, y) / 2) == "dot(B, y)"

    with pytest.raises(TypeError):
        DotProduct(x, y) / B

    assert str(DotProduct(x, y) + 1) == "dot(x, y) + 1"
    assert str(1 + DotProduct(x, y)) == "dot(x, y) + 1"
    assert str(DotProduct(x, y) + DotProduct(B, y)) == "dot(x, y) + dot(B, y)"
    assert str(DotProduct(x, y) + A) == "dot(x, y) + A"
    assert str(B + DotProduct(x, y)) == "dot(x, y) + B"

    assert str(-DotProduct(x, y)) == "-dot(x, y)"
    assert str(-(-DotProduct(x, y))) == "dot(x, y)"

    assert str(DotProduct(x, y) - 1) == "dot(x, y) + -1"
    assert str(1 - DotProduct(x, y)) == "-dot(x, y) + 1"
    assert str(DotProduct(x, y) - DotProduct(B, y)) == "dot(x, y) + -dot(B, y)"
    assert str(DotProduct(x, y) - A) == "dot(x, y) + -A"
    assert str(B - DotProduct(x, y)) == "-dot(x, y) + B"


def test_condition_list():
    xy = DotProduct(x, y)
    Ay = DotProduct(A, y)

    assert str(xy + Ay) == "dot(x, y) + dot(A, y)"

    assert str(2 * (xy + Ay)) == "2 * dot(x, y) + 2 * dot(A, y)"
    assert str((xy + Ay) * 2) == "2 * dot(x, y) + 2 * dot(A, y)"
    assert str(2 * (xy + Ay) * 2) == "4 * dot(x, y) + 4 * dot(A, y)"

    assert str(2 * (xy + Ay) / 2) == "dot(x, y) + dot(A, y)"
    assert str(2 * (xy + Ay) / 0.5) == "4 * dot(x, y) + 4 * dot(A, y)"

    assert str(xy + Ay + xy) == "dot(x, y) + dot(A, y) + dot(x, y)"
    assert str((xy + Ay) + xy) == "dot(x, y) + dot(A, y) + dot(x, y)"
    assert str(2 * (xy + Ay) + xy) == "2 * dot(x, y) + 2 * dot(A, y) + dot(x, y)"
    assert str(xy + (Ay + xy)) == "dot(x, y) + dot(A, y) + dot(x, y)"

    assert str(xy + (Ay - xy)) == "dot(x, y) + dot(A, y) + -dot(x, y)"
    assert str((xy + Ay) - xy) == "dot(x, y) + dot(A, y) + -dot(x, y)"
    assert str(xy - (Ay + xy)) == "dot(x, y) + -dot(A, y) + -dot(x, y)"

    assert str((xy + Ay) - 1) == "dot(x, y) + dot(A, y) + -1"
    assert str((xy + Ay) + 1) == "dot(x, y) + dot(A, y) + 1"
    assert str(1 - (xy + Ay)) == "-dot(x, y) + -dot(A, y) + 1"
    assert str(1 + (xy + Ay)) == "dot(x, y) + dot(A, y) + 1"

    assert str((xy + Ay) + A) == "dot(x, y) + dot(A, y) + A"


def test_combined_source():
    assert str(x * y) == "((x) * (y)) * 1"
    assert str(x * 2 * y) == "((2 * x) * (y)) * 1"
    assert str((x * y) * 2) == "((x) * (y)) * 2"
    assert str((x * y) * A) == "((x) * (y)) * A"
    assert str((-x * A) * y) == "(((-1 * A) * x) * (y)) * 1"
    assert str(x * y + x) == "((x) * (y)) * 1 + x"
    assert str(x * y + y * x) == "((x) * (y)) * 1 + ((y) * (x)) * 1"


def test_source_addition():
    assert str(x + y) == "x + y"
    assert str(x + A) == "x + A"
    assert str(A + x) == "x + A"
    assert str(x + 1) == "x + 1"
    assert str(1 + x) == "x + 1"

    with pytest.raises(TypeError):
        x + []
    with pytest.raises(TypeError):
        x + "1"

    assert str(x - y) == "x + -1 * y"
    assert str(x - A) == "x + -A"
    assert str(A - x) == "-1 * x + A"
    assert str(1 - x) == "-1 * x + 1"
    assert str(x - 1) == "x + -1"

    with pytest.raises(TypeError):
        x - {}
    with pytest.raises(TypeError):
        "1" - x


def test_vector_list():
    xy = x + y

    assert str(xy * 2) == "2 * x + 2 * y"
    assert str((xy * 2) * 2) == "(2 * 2) * x + (2 * 2) * y"
    assert str((xy * 2) / 0.5) == "(2 * 2) * x + (2 * 2) * y"

    assert str(2 * xy) == "2 * x + 2 * y"
    assert str(2 * (xy * 2)) == "(2 * 2) * x + (2 * 2) * y"
    assert str(2 * (2 * xy)) == "(2 * 2) * x + (2 * 2) * y"

    assert str(xy + x) == "x + y + x"
    assert str(xy + 0.5) == "x + y + 0.5"
    assert str(xy + A) == "x + y + A"
    assert str(xy + xy) == "x + y + x + y"
    assert str(xy + 2 * xy) == "x + y + 2 * x + 2 * y"

    with pytest.raises(TypeError):
        xy + ""

    assert str(x + xy) == "x + x + y"
    assert str(0.5 + xy) == "x + y + 0.5"
    assert str(A + xy) == "x + y + A"

    with pytest.raises(TypeError):
        "" + xy

    assert str(xy - x) == "x + y + -1 * x"
    assert str(xy - 0.5) == "x + y + -0.5"
    assert str(xy - A) == "x + y + -A"
    assert str(xy - xy) == "x + y + -1 * x + -1 * y"
    assert str(xy - 2 * xy) == "x + y + -2 * x + -2 * y"

    with pytest.raises(TypeError):
        xy - ""

    assert str(x - xy) == "x + -1 * x + -1 * y"
    assert str(0.5 - xy) == "-1 * x + -1 * y + 0.5"
    assert str(A - xy) == "-1 * x + -1 * y + A"

    with pytest.raises(TypeError):
        "" - xy

    assert str(-xy) == "-1 * x + -1 * y"
    assert str(-(-xy)) == "x + y"



import itertools
import sys
import threading

import numpy as np


def signals_allclose(  # noqa: C901
    t,
    targets,
    signals,
    atol=1e-8,
    rtol=1e-5,
    buf=0,
    delay=0,
    plt=None,
    labels=None,
    individual_results=False,
    allclose=np.allclose,
):
    """Ensure all signal elements are within tolerances.

    Allows for delay, removing the beginning of the signal, and plotting.

    Parameters
    ----------
    t : array_like (T,)
        Simulation time for the points in ``target`` and ``signals``.
    targets : array_like (T, 1) or (T, N)
        Reference signal or signals for error comparison.
    signals : array_like (T, N)
        Signals to be tested against the target signals.
    atol, rtol : float
        Absolute and relative tolerances.
    buf : float
        Length of time (in seconds) to remove from the beginnings of signals.
    delay : float
        Amount of delay (in seconds) to account for when doing comparisons.
    plt : matplotlib.pyplot or mock
        Pyplot interface for plotting the results, unless it's mocked out.
    labels : list of string, length N
        Labels of each signal to use when plotting.
    individual_results : bool
        If True, returns a separate ``allclose`` result for each signal.
    allclose : callable
        Function to compare two arrays for similarity.
    """
    t = np.asarray(t)
    dt = t[1] - t[0]
    assert t.ndim == 1
    assert np.allclose(np.diff(t), dt)  # always use default allclose here

    targets = np.asarray(targets)
    signals = np.asarray(signals)
    if targets.ndim == 1:
        targets = targets.reshape((-1, 1))
    if signals.ndim == 1:
        signals = signals.reshape((-1, 1))
    assert targets.ndim == 2 and signals.ndim == 2
    assert t.size == targets.shape[0]
    assert t.size == signals.shape[0]
    assert targets.shape[1] == 1 or targets.shape[1] == signals.shape[1]

    buf = int(np.round(buf / dt))
    delay = int(np.round(delay / dt))
    slice1 = slice(buf, len(t) - delay)
    slice2 = slice(buf + delay, None)

    if plt is not None:
        if labels is None:
            labels = [None] * len(signals)
        elif isinstance(labels, str):
            labels = [labels]

        colors = ["b", "g", "r", "c", "m", "y", "k"]

        def plot_target(ax, x, b=0, c="k"):
            bound = atol + rtol * np.abs(x)
            y = x - b
            ax.plot(t[slice2], y[slice1], c + ":")
            ax.plot(t[slice2], (y + bound)[slice1], c + "--")
            ax.plot(t[slice2], (y - bound)[slice1], c + "--")

        # signal plot
        ax = plt.subplot(2, 1, 1)
        for y, label in zip(signals.T, labels):
            ax.plot(t, y, label=label)

        if targets.shape[1] == 1:
            plot_target(ax, targets[:, 0], c="k")
        else:
            color_cycle = itertools.cycle(colors)
            for x in targets.T:
                plot_target(ax, x, c=next(color_cycle))

        ax.set_ylabel("signal")
        if labels[0] is not None:
            lgd = ax.legend(loc="upper left", bbox_to_anchor=(1.0, 1.0))
            plt.bbox_extra_artists = (lgd,)

        ax = plt.subplot(2, 1, 2)
        if targets.shape[1] == 1:
            x = targets[:, 0]
            plot_target(ax, x, b=x, c="k")
            for y, label in zip(signals.T, labels):
                ax.plot(t[slice2], y[slice2] - x[slice1])
        else:
            color_cycle = itertools.cycle(colors)
            for x, y, label in zip(targets.T, signals.T, labels):
                c = next(color_cycle)
                plot_target(ax, x, b=x, c=c)
                ax.plot(t[slice2], y[slice2] - x[slice1], c, label=label)

        ax.set_xlabel("time")
        ax.set_ylabel("error")

    if individual_results:
        if targets.shape[1] == 1:
            return [
                allclose(y[slice2], targets[slice1, 0], atol=atol, rtol=rtol)
                for y in signals.T
            ]
        else:
            return [
                allclose(y[slice2], x[slice1], atol=atol, rtol=rtol)
                for x, y in zip(targets.T, signals.T)
            ]
    else:
        return allclose(signals[slice2, :], targets[slice1, :], atol=atol, rtol=rtol)


class ThreadedAssertion:
    """Performs assertions in parallel.

    Starts a number of threads, waits for each thread to execute some
    initialization code, and then executes assertions in each thread.

    To use this class, create a derived class that implements ``init_thread``
    to start each thread running and ``assert_thread`` to check that the thread
    has run successfully. ``finish_thread`` can be used for any cleanup/shutdown
    of the thread.
    """

    class AssertionWorker(threading.Thread):
        def __init__(self, parent, barriers, n):
            super().__init__()
            self.parent = parent
            self.barriers = barriers
            self.n = n
            self.exception = None
            self.exc_info = (None, None, None)

        def run(self):
            self.parent.init_thread(self)

            self.barriers[self.n].set()
            for barrier in self.barriers:
                barrier.wait()

            try:
                self.parent.assert_thread(self)
            except Exception as e:  # pylint: disable=broad-except
                self.exception = e
                self.exc_info = sys.exc_info()
            finally:
                self.parent.finish_thread(self)

    def __init__(self, n_threads):
        self.barriers = [threading.Event() for _ in range(n_threads)]
        self.threads = [
            self.AssertionWorker(self, self.barriers, i) for i in range(n_threads)
        ]

    def run(self):
        for t in self.threads:
            t.start()
        for t in self.threads:
            t.join()
            if t.exception is not None:
                raise t.exception

    def init_thread(self, worker):
        pass

    def assert_thread(self, worker):
        raise NotImplementedError()

    def finish_thread(self, worker):
        pass



import warnings

import numpy as np

from nengo.config import Config
from nengo.connection import Connection
from nengo.dists import Choice, Exponential, Uniform
from nengo.ensemble import Ensemble
from nengo.exceptions import ValidationError
from nengo.network import Network
from nengo.node import Node
from nengo.utils.network import with_self
from nengo.utils.numpy import is_iterable

from .ensemblearray import EnsembleArray


def filtered_step(t, shift=0.5, scale=50, step_val=1):
    return np.maximum(-1 / np.exp((t - shift) * scale) + 1, 0) * step_val


class AssociativeMemory(Network):
    """Associative memory network.

    Parameters
    ----------
    input_vectors: array_like
        The list of vectors to be compared against.
    output_vectors: array_like, optional
        The list of vectors to be produced for each match. If None, the
        associative memory will be autoassociative (cleanup memory).
    n_neurons: int, optional
        The number of neurons for each of the ensemble (where each ensemble
        represents each item in the input_vectors list).
    threshold: float, optional
        The association activation threshold.
    input_scales: float or array_like, optional
        Scaling factor to apply on each of the input vectors. Note that it
        is possible to scale each vector independently.
    inhibitable: bool, optional
        Flag to indicate if the entire associative memory module is
        inhibitable (entire thing can be shut off). The input gain into
        the inhibitory connection is 1.5.
    label : str, optional
        A name for the ensemble. Used for debugging and visualization.
    seed : int, optional
        The seed used for random number generation.
    add_to_container : bool, optional
        Determines if the network will be added to the current container.
        If None, will be true if currently within a Network.
    """

    exp_scale = 0.15  # Scaling factor for exponential distribution
    n_eval_points = 5000

    def __init__(  # noqa: C901
        self,
        input_vectors,
        output_vectors=None,
        n_neurons=50,
        threshold=0.3,
        input_scales=1.0,
        inhibitable=False,
        label=None,
        seed=None,
        add_to_container=None,
    ):
        super().__init__(label, seed, add_to_container)

        # --- Put arguments in canonical form
        if output_vectors is None:
            # If output vocabulary is not specified, use input vector list
            # (i.e autoassociative memory)
            output_vectors = input_vectors
        if is_iterable(input_vectors):
            input_vectors = np.array(input_vectors, ndmin=2)
        if is_iterable(output_vectors):
            output_vectors = np.array(output_vectors, ndmin=2)

        if input_vectors.shape[0] == 0:
            raise ValidationError(
                "Number of input vectors cannot be 0.", attr="input_vectors", obj=self
            )
        elif input_vectors.shape[0] != output_vectors.shape[0]:
            # Fail if number of input items and number of output items don't
            # match
            raise ValidationError(
                "Number of input vectors does not match number of output "
                f"vectors. {input_vectors.shape[0]} != {output_vectors.shape[0]}",
                attr="input_vectors",
                obj=type(self),
            )

        # Handle possible different threshold / input_scale values for each
        # element in the associative memory
        if not is_iterable(threshold):
            threshold = threshold * np.ones(input_vectors.shape[0])
        else:
            threshold = np.array(threshold)

        # --- Check preconditions
        self.n_items = input_vectors.shape[0]
        if threshold.shape[0] != self.n_items:
            raise ValidationError(
                f"Number of threshold values ({threshold.shape[0]}) does not match "
                f"number of input vectors ({self.n_items}).",
                attr="threshold",
                obj=self,
            )

        # --- Set parameters
        self.out_conns = []  # Used in `add_threshold_to_output`
        # Used in `add_threshold_to_output`
        self.default_vector_inhibit_conns = []
        self.thresh_ens = None  # Will hold thresholded outputs
        self.is_wta = False
        self._inhib_scale = 1.5

        # -- Create the core network
        with self, self.am_ens_config:
            self.bias_node = Node(output=1)
            self.elem_input = Node(size_in=self.n_items, label="element input")
            self.elem_output = Node(size_in=self.n_items, label="element output")
            self.utilities = self.elem_output

            self.am_ensembles = []
            label_prefix = "" if label is None else label + "_"
            filt_scale = 15
            filt_step_func = lambda x: filtered_step(x, 0.0, scale=filt_scale)
            for i in range(self.n_items):
                e = Ensemble(n_neurons, 1, label=label_prefix + str(i))
                self.am_ensembles.append(e)

                # Connect input and output nodes
                Connection(self.bias_node, e, transform=-threshold[i])
                Connection(self.elem_input[i], e)
                Connection(e, self.elem_output[i], function=filt_step_func)

            if inhibitable:
                # Input node for inhibitory gating signal (if enabled)
                self.inhibit = Node(size_in=1, label="inhibit")
                Connection(
                    self.inhibit,
                    self.elem_input,
                    transform=-np.ones((self.n_items, 1)) * self._inhib_scale,
                )
                # Note: We can use a decoded connection here because all the
                # am_ensembles have [1] encoders
            else:
                self.inhibit = None

            self.thresh_bias = None
            self.thresholded_utilities = None

        self.add_input_mapping("input", input_vectors, input_scales)
        self.add_output_mapping("output", output_vectors)

    @property
    def am_ens_config(self):
        """(Config) Defaults for associative memory ensemble creation."""
        cfg = Config(Ensemble, Connection)
        cfg[Ensemble].update(
            {
                "radius": 1,
                "intercepts": Exponential(self.exp_scale, 0.0, 1.0),
                "encoders": Choice([[1]]),
                "eval_points": Uniform(0.0, 1.0),
                "n_eval_points": self.n_eval_points,
            }
        )
        cfg[Connection].synapse = None
        return cfg

    @property
    def default_ens_config(self):
        """(Config) Defaults for other ensemble creation."""
        cfg = Config(Ensemble)
        cfg[Ensemble].update(
            {
                "radius": 1,
                "intercepts": Exponential(self.exp_scale, 0.0, 1.0),
                "encoders": Choice([[1]]),
                "eval_points": Uniform(0.0, 1.0),
                "n_eval_points": self.n_eval_points,
            }
        )
        return cfg

    @property
    def thresh_ens_config(self):
        """(Config) Defaults for threshold ensemble creation."""
        cfg = Config(Ensemble)
        cfg[Ensemble].update(
            {
                "radius": 1,
                "intercepts": Uniform(0.5, 1.0),
                "encoders": Choice([[1]]),
                "eval_points": Uniform(0.75, 1.1),
                "n_eval_points": self.n_eval_points,
            }
        )
        return cfg

    @with_self
    def add_input_mapping(self, name, input_vectors, input_scales=1.0):
        """Adds a set of input vectors to the associative memory network.

        Creates a transform with the given input vectors between the
        a named input node and associative memory element input to enable the
        inputs to be mapped onto ensembles of the Associative Memory.

        Parameters
        ----------
        name: str
            Name to use for the input node. This name will be used as the name
            of the attribute for the associative memory network.
        input_vectors: array_like
            The list of vectors to be compared against.
        input_scales: float or array_like, optional
            Scaling factor to apply on each of the input vectors. Note that it
            is possible to scale each vector independently.
        """
        # --- Put arguments in canonical form
        n_vectors, d_vectors = input_vectors.shape
        if is_iterable(input_vectors):
            input_vectors = np.array(input_vectors, ndmin=2)
        if not is_iterable(input_scales):
            input_scales = input_scales * np.ones((1, n_vectors))
        else:
            input_scales = np.array(input_scales, ndmin=2)

        # --- Check some preconditions
        if input_scales.shape[1] != n_vectors:
            raise ValidationError(
                f"Number of input_scale values ({input_scales.shape[1]}) does not "
                f"match number of input vectors ({n_vectors}).",
                attr="input_scales",
            )
        if hasattr(self, name):
            raise ValidationError(
                f"Name '{name}' already exists as a node in the associative memory.",
                attr="name",
            )

        # --- Finally, make the input node and connect it
        in_node = Node(size_in=d_vectors, label=name)
        setattr(self, name, in_node)
        Connection(
            in_node,
            self.elem_input,
            synapse=None,
            transform=input_vectors * input_scales.T,
        )

    @with_self
    def add_output_mapping(self, name, output_vectors):
        """Adds another output to the associative memory network.

        Creates a transform with the given output vectors between the
        associative memory element output and a named output node to enable the
        selection of output vectors by the associative memory.

        Parameters
        ----------
        name: str
            Name to use for the output node. This name will be used as
            the name of the attribute for the associative memory network.
        output_vectors: array_like
            The list of vectors to be produced for each match.
        """
        # --- Put arguments in canonical form
        if is_iterable(output_vectors):
            output_vectors = np.array(output_vectors, ndmin=2)

        # --- Check preconditions
        if hasattr(self, name):
            raise ValidationError(
                f"Name '{name}' already exists as a node in the associative memory.",
                attr="name",
            )

        # --- Make the output node and connect it
        output = Node(size_in=output_vectors.shape[1], label=name)
        setattr(self, name, output)

        if self.thresh_ens is not None:
            c = Connection(
                self.thresh_ens.output, output, synapse=None, transform=output_vectors.T
            )
        else:
            c = Connection(
                self.elem_output, output, synapse=None, transform=output_vectors.T
            )
        self.out_conns.append(c)

    @with_self
    def add_default_output_vector(
        self,
        output_vector,
        output_name="output",
        n_neurons=50,
        min_activation_value=0.5,
    ):
        """Adds a default output vector to the associative memory network.

        The default output vector is chosen if the input matches none of
        the given input vectors.

        Parameters
        ----------
        output_vector: array_like
            The vector to be produced if the input value matches none of
            the vectors in the input vector list.
        output_name: str, optional
            The name of the input to which the default output vector
            should be applied.
        n_neurons: int, optional
            Number of neurons to use for the default output vector ensemble.
        min_activation_value: float, optional
            Minimum activation value (i.e. threshold) to use to disable
            the default output vector.
        """
        with self.default_ens_config:
            default_vector_ens = Ensemble(
                n_neurons, 1, label=f"Default {output_name} vector"
            )
            Connection(self.bias_node, default_vector_ens, synapse=None)

            tr = -(1.0 / min_activation_value) * np.ones((1, self.n_items))
            if self.thresh_ens is not None:
                c = Connection(self.thresh_ens.output, default_vector_ens, transform=tr)
            else:
                c = Connection(self.elem_output, default_vector_ens, transform=tr)

            # Add the output connection to the output connection list
            self.default_vector_inhibit_conns.append(c)

            # Make new output class attribute and connect to it
            output = getattr(self, output_name)
            Connection(
                default_vector_ens,
                output,
                transform=np.array(output_vector, ndmin=2).T,
                synapse=None,
            )

            if self.inhibit is not None:
                Connection(
                    self.inhibit,
                    default_vector_ens,
                    transform=-self._inhib_scale,
                    synapse=None,
                )

    @with_self
    def add_wta_network(self, inhibit_scale=1.5, inhibit_synapse=0.005):
        """Add a winner-take-all (WTA) network to associative memory output.

        Parameters
        ----------
        inhibit_scale: float, optional
            Mutual inhibition scaling factor.
        inhibit_synapse: float, optional
            Mutual inhibition synapse time constant.
        """
        if not self.is_wta:
            Connection(
                self.elem_output,
                self.elem_input,
                synapse=inhibit_synapse,
                transform=((np.eye(self.n_items) - 1) * inhibit_scale),
            )
            self.is_wta = True
        else:
            warnings.warn(
                "AssociativeMemory network is already configured "
                "with a WTA network. Additional `add_wta_network` "
                "calls are ignored."
            )

    @with_self
    def add_threshold_to_outputs(self, n_neurons=50, inhibit_scale=10):
        """Adds a thresholded output to the associative memory.

        Parameters
        ----------
        n_neurons: int, optional
            Number of neurons to use for the default output vector ensemble.
        inhibit_scale: float, optional
            Mutual inhibition scaling factor.
        """
        if self.thresh_ens is not None:
            warnings.warn(
                "AssociativeMemory network is already configured "
                "with thresholded outputs. Additional "
                "`add_threshold_to_output` calls are ignored."
            )
            return

        with self.thresh_ens_config:
            self.thresh_bias = EnsembleArray(
                n_neurons, self.n_items, label="thresh_bias"
            )
            self.thresh_ens = EnsembleArray(n_neurons, self.n_items, label="thresh_ens")
            self.thresholded_utilities = self.thresh_ens.output

            Connection(
                self.bias_node,
                self.thresh_bias.input,
                transform=np.ones((self.n_items, 1)),
                synapse=None,
            )
            Connection(
                self.bias_node,
                self.thresh_ens.input,
                transform=np.ones((self.n_items, 1)),
                synapse=None,
            )
            Connection(
                self.elem_output, self.thresh_bias.input, transform=-inhibit_scale
            )
            Connection(
                self.thresh_bias.output, self.thresh_ens.input, transform=-inhibit_scale
            )

            # Reroute connections from elem_output to be connections
            # from thresh_ens.output
            def reroute(conn):
                c = Connection(
                    self.thresh_ens.output,
                    conn.post,
                    transform=conn.transform,
                    synapse=conn.synapse,
                )
                self.connections.remove(conn)
                return c

            self.default_vector_inhibit_conns = [
                reroute(c) for c in self.default_vector_inhibit_conns
            ]
            self.out_conns = [reroute(c) for c in self.out_conns]

            # Make inhibitory connection if inhibit option is set
            if self.inhibit is not None:
                for e in self.thresh_ens.ensembles:
                    Connection(
                        self.inhibit, e, transform=-self._inhib_scale, synapse=None
                    )



import inspect
from collections import namedtuple

import numpy as np

from nengo.exceptions import ConfigError, ObsoleteError, ReadonlyError, ValidationError
from nengo.rc import rc
from nengo.utils.numpy import (
    array_hash,
    compare,
    is_array,
    is_array_like,
    is_integer,
    is_number,
)
from nengo.utils.stdlib import WeakKeyIDDictionary, checked_call


class DefaultType:
    """Placeholder object used to represent default values for a parameter."""

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return self.name


Default = DefaultType("Default")
ConnectionDefault = DefaultType("ConnectionDefault")
Unconfigurable = DefaultType("Unconfigurable")


def is_param(obj):
    """Check if ``obj`` is a Parameter."""
    return isinstance(obj, Parameter)


def iter_params(obj):
    """Iterate over the names of all parameters of an object."""
    obj = obj if inspect.isclass(obj) else type(obj)
    return (
        name
        for name in dir(obj)
        if is_param(getattr(obj, name))
        and not isinstance(getattr(obj, name), ObsoleteParam)
    )


def equal(a, b):
    """Check if two (possibly array-like) objects are equal."""
    if isinstance(a, dict) and isinstance(b, dict):
        return set(a) == set(b) and all(equal(a[key], b[key]) for key in a)

    if is_array_like(a) or is_array_like(b):
        return np.array_equal(a, b)

    return a == b


class Parameter:
    """Simple descriptor for storing configuration parameters.

    Parameters
    ----------
    name : str
        Name of the parameter.
    default : object
        The value returned if the parameter hasn't been explicitly set.
    optional : bool, optional
        Whether this parameter accepts the value None. By default,
        parameters are not optional (i.e., cannot be set to ``None``).
    readonly : bool, optional
        If true, the parameter can only be set once.
        By default, parameters can be set multiple times.

    Attributes
    ----------
    coerce_defaults : bool
        If True, validate values for this parameter when they are set in a
        `.Config` object. Setting a parameter directly on an object will
        always be validated.
    equatable : bool
        If True, parameter values can be compared for equality
        (``a==b``); otherwise equality checks will just compare object
        identity (``a is b``).
    """

    coerce_defaults = True
    equatable = False

    def __init__(self, name, default=Unconfigurable, optional=False, readonly=None):
        # freeze Unconfigurables by default
        readonly = default is Unconfigurable if readonly is None else readonly

        if not isinstance(name, str):
            raise ValueError(f"'name' must be a string (got {name!r})")
        if not isinstance(optional, bool):
            raise ValueError(f"'optional' must be boolean (got {optional!r})")
        if not isinstance(readonly, bool):
            raise ValueError(f"'readonly' must be boolean (got {readonly!r})")

        self.name = name
        self.default = default
        self.optional = optional
        self.readonly = readonly

        # default values set by config system
        self._defaults = WeakKeyIDDictionary()

        # param values set on objects
        self.data = WeakKeyIDDictionary()

    def __getstate__(self):
        state = {}
        state.update(self.__dict__)
        state["_defaults"] = dict(state["_defaults"].items())
        state["data"] = dict(state["data"].items())
        return state

    def __setstate__(self, state):
        for k, v in state.items():
            if k in ["_defaults", "data"]:
                v = WeakKeyIDDictionary(v)
            setattr(self, k, v)

    def __contains__(self, key):
        return key in self.data or key in self._defaults

    def __delete__(self, instance):
        del self.data[instance]

    def __get__(self, instance, type_):
        if instance is None:
            # Return self so default can be inspected
            return self

        try:
            return self.data[instance]
        except KeyError as e:
            if not self.configurable:
                raise ValidationError(
                    "Unconfigurable parameters have no defaults. Please ensure the"
                    " value of the parameter is set before trying to access it.",
                    attr=self.name,
                    obj=instance,
                ) from e
            return self.default

    def __set__(self, instance, value):
        self.data[instance] = self.coerce(instance, value)

    def __repr__(self):
        return (
            f"{type(self).__name__}('{self.name}', default={self.default}, "
            f"optional={self.optional}, readonly={self.readonly})"
        )

    @property
    def configurable(self):
        return self.default is not Unconfigurable

    def del_default(self, obj):
        del self._defaults[obj]

    def get_default(self, obj):
        return self._defaults.get(obj, self.default)

    def set_default(self, obj, value):
        if not self.configurable:
            raise ConfigError(f"Parameter '{self}' is not configurable")
        self._defaults[obj] = self.coerce(obj, value) if self.coerce_defaults else value

    def check_type(self, instance, value, type_):
        if value is not None and not isinstance(value, type_):
            if isinstance(type_, tuple):
                type_str = " or ".join((t.__name__ for t in type_))
            else:
                type_str = type_.__name__
            raise ValidationError(
                f"Must be of type '{type_str}' (got type '{type(value).__name__}').",
                attr=self.name,
                obj=instance,
            )

    def coerce(self, instance, value):
        if isinstance(value, DefaultType):
            raise ValidationError(
                "Default is not a valid value. To reset a parameter, use 'del'.",
                attr=self.name,
                obj=instance,
            )
        if self.readonly and instance in self.data:
            raise ReadonlyError(attr=self.name, obj=instance)
        if not self.optional and value is None:
            raise ValidationError(
                "Parameter is not optional; cannot set to None",
                attr=self.name,
                obj=instance,
            )
        return value

    def equal(self, instance_a, instance_b):
        a = self.__get__(instance_a, None)
        b = self.__get__(instance_b, None)
        if self.equatable:
            return equal(a, b)
        else:
            return a is b

    def hashvalue(self, instance):
        """Returns a hashable value (`hash` can be called on the output)."""
        value = self.__get__(instance, None)
        if self.equatable:
            return value
        else:
            return id(value)


class ObsoleteParam(Parameter):
    """A parameter that is no longer supported."""

    def __init__(self, name, short_msg, since=None, url=None):
        self.short_msg = short_msg
        self.since = since
        self.url = url
        super().__init__(name, optional=True)

    def __get__(self, instance, type_):
        if instance is not None:
            self.raise_error()

        # Return self so default can be inspected
        return self

    def coerce(self, instance, value):
        if value is not Unconfigurable:
            # don't allow setting to anything other than unconfigurable default
            self.raise_error()
        return value

    def raise_error(self):
        raise ObsoleteError(self.short_msg, since=self.since, url=self.url)


class BoolParam(Parameter):
    """A parameter where the value is a boolean."""

    equatable = True

    def coerce(self, instance, value):
        self.check_type(instance, value, bool)
        return super().coerce(instance, value)


class NumberParam(Parameter):
    """A parameter where the value is a number."""

    equatable = True

    def __init__(
        self,
        name,
        default=Unconfigurable,
        low=None,
        high=None,
        low_open=False,
        high_open=False,
        optional=False,
        readonly=None,
    ):
        self.low = low
        self.high = high
        self.low_open = low_open
        self.high_open = high_open
        super().__init__(name, default, optional, readonly)

    def coerce(self, instance, num):  # pylint: disable=arguments-renamed
        if num is not None:
            if is_array(num) and num.shape == ():
                num = num.item()  # convert scalar array to Python object

            if not is_number(num):
                raise ValidationError(
                    f"Must be a number; got '{num}'", attr=self.name, obj=instance
                )
            low_comp = 0 if self.low_open else -1
            if self.low is not None and compare(num, self.low) <= low_comp:
                eq_phrase = "" if self.low_open else " or equal to"
                raise ValidationError(
                    f"Value must be greater than{eq_phrase} {self.low} (got {num})",
                    attr=self.name,
                    obj=instance,
                )
            high_comp = 0 if self.high_open else 1
            if self.high is not None and compare(num, self.high) >= high_comp:
                eq_phrase = "" if self.high_open else " or equal to"
                raise ValidationError(
                    f"Value must be less than{eq_phrase} {self.high} (got {num})",
                    attr=self.name,
                    obj=instance,
                )
        return super().coerce(instance, num)


class IntParam(NumberParam):
    """A parameter where the value is an integer."""

    def coerce(self, instance, num):
        self.check_type(instance, num, (int, np.integer))
        return super().coerce(instance, num)


class StringParam(Parameter):
    """A parameter where the value is a string."""

    equatable = True

    def coerce(self, instance, string):  # pylint: disable=arguments-renamed
        self.check_type(instance, string, (str,))
        return super().coerce(instance, string)


class EnumParam(StringParam):
    """A parameter where the value must be one of a finite set of strings."""

    def __init__(
        self,
        name,
        default=Unconfigurable,
        values=(),
        lower=True,
        optional=False,
        readonly=None,
    ):
        assert all(isinstance(s, str) for s in values)
        if lower:
            values = tuple(s.lower() for s in values)
        value_set = set(values)
        assert len(values) == len(value_set)
        self.values = values
        self.value_set = value_set
        self.lower = lower
        super().__init__(name, default, optional, readonly)

    def coerce(self, instance, string):
        string = super().coerce(instance, string)
        string = string.lower() if self.lower else string
        if string not in self.value_set:
            raise ValidationError(
                f"String '{string}' must be one of {list(self.values)}",
                attr=self.name,
                obj=instance,
            )
        return string


class TupleParam(Parameter):
    """A parameter where the value is a tuple."""

    def __init__(
        self, name, default=Unconfigurable, length=None, optional=False, readonly=None
    ):
        self.length = length
        super().__init__(name, default, optional, readonly)

    def coerce(self, instance, value):
        value = super().coerce(instance, value)

        if value is not None:
            try:
                value = tuple(value)
            except TypeError as e:
                raise ValidationError(
                    "Value must be castable to a tuple", attr=self.name, obj=instance
                ) from e

            if self.length is not None and len(value) != self.length:
                raise ValidationError(
                    f"Must be {self.length} items (got {len(value)})",
                    attr=self.name,
                    obj=instance,
                )
        return value


class ShapeParam(TupleParam):
    """A parameter where the value is a tuple of integers."""

    equatable = True

    def __init__(
        self,
        name,
        default=Unconfigurable,
        length=None,
        low=0,
        optional=False,
        readonly=None,
    ):
        super().__init__(
            name, default=default, length=length, optional=optional, readonly=readonly
        )
        self.low = low

    def coerce(self, instance, value):
        value = super().coerce(instance, value)
        if value is not None:
            for i, v in enumerate(value):
                if not is_integer(v):
                    raise ValidationError(
                        f"Element {i} must be an int (got type {type(v).__name__})",
                        attr=self.name,
                        obj=instance,
                    )
                if self.low is not None and v < self.low:
                    raise ValidationError(
                        f"Element {i} must be >= {self.low} (got {v})",
                        attr=self.name,
                        obj=instance,
                    )

        return value


class DictParam(Parameter):
    """A parameter where the value is a dictionary."""

    equatable = True

    def coerce(self, instance, value):
        self.check_type(instance, value, dict)
        return super().coerce(instance, value)

    def hashvalue(self, instance):
        d = self.__get__(instance, None)
        if d is None:
            return hash(d)
        return hash(
            tuple(
                (k, array_hash(v) if is_array_like(v) else hash(v))
                for k, v in d.items()
            )
        )


class NdarrayParam(Parameter):
    """A parameter where the value is a NumPy ndarray.

    If the passed value is an ndarray, a view onto that array is stored.
    If the passed value is not an ndarray, it will be cast to an ndarray
    of ``dtype`` and stored.
    """

    equatable = True

    def __init__(
        self,
        name,
        default=Unconfigurable,
        shape=None,
        dtype=None,
        optional=False,
        readonly=None,
    ):
        if shape is not None:
            assert shape.count("...") <= 1, "Cannot have more than one ellipsis"
        self.shape = shape
        self._dtype = dtype
        super().__init__(name, default, optional, readonly)

    @property
    def coerce_defaults(self):
        if self.shape is None:
            return True
        return all(is_integer(dim) or dim in ("...", "*") for dim in self.shape)

    @property
    def dtype(self):
        if self._dtype is not None:
            return self._dtype
        return rc.float_dtype

    def hashvalue(self, instance):
        return array_hash(self.__get__(instance, None))

    def coerce(self, instance, value):
        if value is not None:
            value = self.coerce_ndarray(instance, value)
        return super().coerce(instance, value)

    def coerce_ndarray(self, instance, ndarray):  # noqa: C901
        if isinstance(ndarray, np.ndarray):
            ndarray = ndarray.view()
        else:
            try:
                ndarray = np.array(ndarray, dtype=self.dtype)
            except (ValueError, TypeError) as e:
                raise ValidationError(
                    f"Must be a {self.dtype} NumPy array "
                    f"(got type {type(ndarray).__name__})",
                    attr=self.name,
                    obj=instance,
                ) from e

        if self.readonly:
            ndarray.setflags(write=False)

        if self.shape is None:
            return ndarray

        if "..." in self.shape:
            # Convert '...' to the appropriate number of '*'s
            nfixed = len(self.shape) - 1
            n = ndarray.ndim - nfixed
            if n < 0:
                raise ValidationError(
                    f"ndarray must be at least {nfixed}D (got {ndarray.ndim}D)",
                    attr=self.name,
                    obj=instance,
                )

            i = self.shape.index("...")
            shape = list(self.shape[:i]) + (["*"] * n)
            if i < len(self.shape) - 1:
                shape.extend(self.shape[i + 1 :])
        else:
            shape = self.shape

        if ndarray.ndim != len(shape):
            raise ValidationError(
                f"ndarray must be {len(shape)}D (got {ndarray.ndim}D)",
                attr=self.name,
                obj=instance,
            )

        for i, attr in enumerate(shape):
            assert is_integer(attr) or isinstance(
                attr, str
            ), "shape can only be an int or str representing an attribute"
            if attr == "*":
                continue

            desired = attr if is_integer(attr) else getattr(instance, attr)

            if not is_integer(desired):
                raise ValidationError(
                    f"{attr} not yet initialized; cannot determine if shape is "
                    "correct. Consider using a distribution instead.",
                    attr=self.name,
                    obj=instance,
                )

            if ndarray.shape[i] != desired:
                raise ValidationError(
                    f"shape[{i}] should be {desired} (got {ndarray.shape[i]})",
                    attr=self.name,
                    obj=instance,
                )
        return ndarray


FunctionInfo = namedtuple("FunctionInfo", ["function", "size"])


class FunctionParam(Parameter):
    """A parameter where the value is a function."""

    def determine_size(self, instance, function):
        args = self.function_args(instance, function)
        value, invoked = checked_call(function, *args)
        if not invoked:
            raise ValidationError(
                f"function '{function}' must accept a single np.array argument",
                attr=self.name,
                obj=instance,
            )
        return np.asarray(value).size

    def function_args(self, instance, function):
        return (np.zeros(1),)

    def coerce(self, instance, function):  # pylint: disable=arguments-renamed
        function = super().coerce(instance, function)
        if isinstance(function, FunctionInfo):
            function_info = function
            function = function_info.function
        else:
            size = (
                self.determine_size(instance, function) if callable(function) else None
            )
            function_info = FunctionInfo(function=function, size=size)

        if function is not None and not callable(function):
            raise ValidationError(
                f"function '{function}' must be callable",
                attr=self.name,
                obj=instance,
            )

        return function_info


class FrozenObject:
    """An object with parameters that cannot change value after instantiation.

    Since such objects are read-only ("frozen"), they can be safely used in
    multiple locations, compared, etc.
    """

    # Order in which parameters have to be initialized.
    # Missing parameters will be initialized last in an undefined order.
    # This is needed for pickling and copying of Nengo objects when the
    # parameter initialization order matters.
    _param_init_order = []

    def __init__(self):
        self._paramdict = dict(
            (k, v)
            for k, v in inspect.getmembers(type(self))
            if isinstance(v, Parameter) and not isinstance(v, ObsoleteParam)
        )
        for p in self._params:
            if not p.readonly:
                msg = "All parameters of a FrozenObject must be readonly"
                raise ReadonlyError(attr=p, obj=self, msg=msg)
        self.__argreprs = None

    @property
    def _params(self):
        return list(self._paramdict.values())

    def __eq__(self, other):
        if self is other:  # quick check for speed
            return True
        return type(self) == type(other) and all(
            p.equal(self, other) for p in self._params
        )

    def __hash__(self):
        return hash((type(self), tuple(p.hashvalue(self) for p in self._params)))

    def __getstate__(self):
        d = dict(self.__dict__)
        d.pop("_paramdict")  # do not pickle the param dict itself
        for k in self._paramdict:
            d[k] = getattr(self, k)

        return d

    def __setstate__(self, state):
        FrozenObject.__init__(self)  # set up the param dict

        for attr in self._param_init_order:
            setattr(self, attr, state.pop(attr))
        for attr in set(self._paramdict).difference(self._param_init_order):
            setattr(self, attr, state.pop(attr))

        self.__dict__.update(state)

    def __repr__(self):
        if isinstance(self._argreprs, str):
            return f"<{type(self).__name__} at 0x{id(self):x}>"
        return f"{type(self).__name__}({', '.join(self._argreprs)})"

    @property
    def _argreprs(self):
        if self.__argreprs is not None:
            return self.__argreprs

        # get arguments to display from __init__ functions
        spec = inspect.getfullargspec(type(self).__init__)
        defaults = {}
        if spec.defaults is not None:
            defaults.update(zip(spec.args[-len(spec.defaults) :], spec.defaults))

        self.__argreprs = []
        for arg in spec.args[1:]:  # start at 1 to drop `self`
            if not hasattr(self, arg):
                # We rely on storing the initial arguments. If we don't have
                # them, we don't auto-generate a repr.
                self.__argreprs = f"Cannot find '{arg}'"
                break
            value = getattr(self, arg)

            param = self._paramdict.get(arg, None)
            if arg in defaults and defaults[arg] is not Default:
                not_default = not equal(value, defaults[arg])
            elif param is not None and param.default is not Unconfigurable:
                not_default = not equal(value, param.default)
            else:
                not_default = True

            if not_default:
                self.__argreprs.append(f"{arg}={value!r}")

        return self.__argreprs



import warnings

import numpy as np

from nengo.dists import Choice, Distribution, Uniform, get_samples
from nengo.exceptions import SimulationError, ValidationError
from nengo.params import DictParam, FrozenObject, NumberParam, Parameter
from nengo.rc import rc
from nengo.utils.numpy import clip, is_array_like


def settled_firingrate(step, J, state, dt=0.001, settle_time=0.1, sim_time=1.0):
    """Compute firing rates (in Hz) for given vector input, ``x``.

    Unlike the default naive implementation, this approach takes into
    account some characteristics of spiking neurons. We start
    by simulating the neurons for a short amount of time, to let any
    initial transients settle. Then, we run the neurons for a second
    and find the average (which should approximate the firing rate).

    Parameters
    ----------
    step : function
        the step function of the neuron type
    J : ndarray
        a vector of currents to generate firing rates from
    state : dict of ndarrays
        additional state needed by the step function
    """
    total = np.zeros_like(J)
    out = state["output"]

    # Simulate for the settle time
    steps = int(settle_time / dt)
    for _ in range(steps):
        step(dt, J, **state)
    # Simulate for sim time, and keep track
    steps = int(sim_time / dt)
    for _ in range(steps):
        step(dt, J, **state)
        total += out
    return total / float(steps)


class NeuronType(FrozenObject):
    """Base class for Nengo neuron models.

    Parameters
    ----------
    initial_state : {str: Distribution or array_like}
        Mapping from state variables names to their desired initial value.
        These values will override the defaults set in the class's state attribute.

    Attributes
    ----------
    state : {str: Distribution}
        State variables held by the neuron type during simulation.
        Values in the dict indicate their initial values, or how
        to obtain those initial values. These elements can also be
        probed in the neuron population.
    negative : bool
        Whether the neurons can emit negative outputs (i.e. negative spikes or rates).
    """

    state = {}
    negative = True
    spiking = False

    initial_state = DictParam("initial_state", optional=True)

    def __init__(self, initial_state=None):
        super().__init__()
        self.initial_state = initial_state
        if self.initial_state is not None:
            for name, value in self.initial_state.items():
                if name not in self.state:
                    raise ValidationError(
                        f"State variable '{name}' not recognized; should be one of "
                        f"{', '.join(repr(k) for k in self.state)}",
                        attr="initial_state",
                        obj=self,
                    )
                if not (isinstance(value, Distribution) or is_array_like(value)):
                    raise ValidationError(
                        f"State variable '{name}' must be a distribution or array-like",
                        attr="initial_state",
                        obj=self,
                    )

    @property
    def probeable(self):
        return ("output",) + tuple(self.state)

    def current(self, x, gain, bias):
        """Compute current injected in each neuron given input, gain and bias.

        Note that ``x`` is assumed to be already projected onto the encoders
        associated with the neurons and normalized to radius 1, so the maximum
        expected current for a neuron occurs when input for that neuron is 1.

        Parameters
        ----------
        x : (n_samples,) or (n_samples, n_neurons) array_like
            Scalar inputs for which to calculate current.
        gain : (n_neurons,) array_like
            Gains associated with each neuron.
        bias : (n_neurons,) array_like
            Bias current associated with each neuron.

        Returns
        -------
        current : (n_samples, n_neurons)
            Current to be injected in each neuron.
        """
        x = np.array(x, dtype=float, copy=False, ndmin=1)
        gain = np.array(gain, dtype=float, copy=False, ndmin=1)
        bias = np.array(bias, dtype=float, copy=False, ndmin=1)

        if x.ndim == 1:
            x = x[:, np.newaxis]
        elif x.ndim >= 3 or x.shape[1] != gain.shape[0]:
            raise ValidationError(
                f"Expected shape {(x.shape[0], gain.shape[0])}; got {x.shape}.",
                attr="x",
                obj=self,
            )

        return gain * x + bias

    def gain_bias(self, max_rates, intercepts):
        """Compute the gain and bias needed to satisfy max_rates, intercepts.

        This takes the neurons, approximates their response function, and then
        uses that approximation to find the gain and bias value that will give
        the requested intercepts and max_rates.

        Note that this default implementation is very slow! Whenever possible,
        subclasses should override this with a neuron-specific implementation.

        Parameters
        ----------
        max_rates : (n_neurons,) array_like
            Maximum firing rates of neurons.
        intercepts : (n_neurons,) array_like
            X-intercepts of neurons.

        Returns
        -------
        gain : (n_neurons,) array_like
            Gain associated with each neuron. Sometimes denoted alpha.
        bias : (n_neurons,) array_like
            Bias current associated with each neuron.
        """
        max_rates = np.array(max_rates, dtype=float, copy=False, ndmin=1)
        intercepts = np.array(intercepts, dtype=float, copy=False, ndmin=1)

        J_steps = 101  # Odd number so that 0 is a sample
        max_rate = max_rates.max()

        # Start with dummy gain and bias so x == J in rate calculation
        gain = np.ones(1)
        bias = np.zeros(1)

        # Find range of J that will achieve max rates (assume monotonic)
        J_threshold = None
        J_max = None
        Jr = 10
        for _ in range(10):
            J = np.linspace(-Jr, Jr, J_steps)
            rate = self.rates(J, gain, bias)
            if J_threshold is None and (rate <= 0).any():
                J_threshold = J[np.where(rate <= 0)[0][-1]]
            if J_max is None and (rate >= max_rate).any():
                J_max = J[np.where(rate >= max_rate)[0][0]]

            if J_threshold is not None and J_max is not None:
                break

            Jr *= 2
        else:
            if J_threshold is None:
                raise ValidationError(
                    "Could not find firing threshold",
                    attr="max_rates,intercepts",
                    obj=self,
                )
            if J_max is None:
                raise ValidationError(
                    "Could not find max current", attr="max_rates", obj=self
                )

        J = np.linspace(J_threshold, J_max, J_steps)
        rate = self.rates(J, gain, bias).squeeze(axis=1)

        gain = np.zeros_like(max_rates)
        bias = np.zeros_like(max_rates)
        J_tops = np.interp(max_rates, rate, J)

        gain[:] = (J_threshold - J_tops) / (intercepts - 1)
        bias[:] = J_tops - gain
        return gain, bias

    def make_state(self, n_neurons, rng=np.random, dtype=None):
        dtype = rc.float_dtype if dtype is None else dtype
        state = {}
        initial_state = {} if self.initial_state is None else self.initial_state
        for name, default_value in self.state.items():
            dist = initial_state.get(name, default_value)
            state[name] = get_samples(dist, n=n_neurons, d=None, rng=rng).astype(
                dtype, copy=False
            )
        return state

    def max_rates_intercepts(self, gain, bias):
        """Compute the max_rates and intercepts given gain and bias.

        Note that this default implementation is very slow! Whenever possible,
        subclasses should override this with a neuron-specific implementation.

        Parameters
        ----------
        gain : (n_neurons,) array_like
            Gain associated with each neuron. Sometimes denoted alpha.
        bias : (n_neurons,) array_like
            Bias current associated with each neuron.

        Returns
        -------
        max_rates : (n_neurons,) array_like
            Maximum firing rates of neurons.
        intercepts : (n_neurons,) array_like
            X-intercepts of neurons.
        """

        max_rates = self.rates(1, gain, bias).squeeze(axis=0)

        x_range = np.linspace(-1, 1, 101)
        rates = self.rates(x_range, gain, bias)
        last_zeros = np.maximum(np.argmax(rates > 0, axis=0) - 1, 0)
        intercepts = x_range[last_zeros]

        return max_rates, intercepts

    def rates(self, x, gain, bias):
        """Compute firing rates (in Hz) for given input ``x``.

        This default implementation takes the naive approach of running the
        step function for a second. This should suffice for most rate-based
        neuron types; for spiking neurons it will likely fail (those models
        should override this function).

        Note that ``x`` is assumed to be already projected onto the encoders
        associated with the neurons and normalized to radius 1, so the maximum
        expected rate for a neuron occurs when input for that neuron is 1.

        Parameters
        ----------
        x : (n_samples,) or (n_samples, n_neurons) array_like
            Scalar inputs for which to calculate rates.
        gain : (n_neurons,) array_like
            Gains associated with each neuron.
        bias : (n_neurons,) array_like
            Bias current associated with each neuron.

        Returns
        -------
        rates : (n_samples, n_neurons) ndarray
            The firing rates at each given value of ``x``.
        """
        J = self.current(x, gain, bias)
        out = np.zeros_like(J)
        self.step(dt=1.0, J=J, output=out)
        return out

    def step(self, dt, J, output, **state):
        """Implements the differential equation for this neuron type.

        At a minimum, NeuronType subclasses must implement this method.
        That implementation should modify the ``output`` parameter rather
        than returning anything, for efficiency reasons.

        Parameters
        ----------
        dt : float
            Simulation timestep.
        J : (n_neurons,) array_like
            Input currents associated with each neuron.
        output : (n_neurons,) array_like
            Output activity associated with each neuron (e.g., spikes or firing rates).
        state : {str: array_like}
            State variables associated with the population.
        """
        raise NotImplementedError("Neurons must provide step")

    def step_math(self, dt, J, output, **state):
        warnings.warn(
            "'step_math' has been renamed to 'step'. This alias will be removed "
            "in Nengo 4.0"
        )
        return self.step(dt, J, output, **state)


class NeuronTypeParam(Parameter):

    equatable = True

    def coerce(self, instance, neurons):  # pylint: disable=arguments-renamed
        self.check_type(instance, neurons, NeuronType)
        return super().coerce(instance, neurons)


class Direct(NeuronType):
    """Signifies that an ensemble should simulate in direct mode.

    In direct mode, the ensemble represents and transforms signals perfectly,
    rather than through a neural approximation. Note that direct mode ensembles
    with recurrent connections can easily diverge; most other neuron types will
    instead saturate at a certain high firing rate.
    """

    def gain_bias(self, max_rates, intercepts):
        """Always returns ``None, None``."""
        return None, None

    def max_rates_intercepts(self, gain, bias):
        """Always returns ``None, None``."""
        return None, None

    def rates(self, x, gain, bias):
        """Always returns ``x``."""
        return np.array(x, dtype=float, copy=False, ndmin=1)

    def step(self, dt, J, output):
        """Raises an error if called.

        Rather than calling this function, the simulator will detect that
        the ensemble is in direct mode, and bypass the neural approximation.
        """
        raise SimulationError("Direct mode neurons shouldn't be simulated.")


class RectifiedLinear(NeuronType):
    """A rectified linear neuron model.

    Each neuron is modeled as a rectified line. That is, the neuron's activity
    scales linearly with current, unless it passes below zero, at which point
    the neural activity will stay at zero.

    Parameters
    ----------
    amplitude : float
        Scaling factor on the neuron output. Corresponds to the relative
        amplitude of the output of the neuron.
    initial_state : {str: Distribution or array_like}
        Mapping from state variables names to their desired initial value.
        These values will override the defaults set in the class's state attribute.
    """

    negative = False

    amplitude = NumberParam("amplitude", low=0, low_open=True)

    def __init__(self, amplitude=1, initial_state=None):
        super().__init__(initial_state)

        self.amplitude = amplitude

    def gain_bias(self, max_rates, intercepts):
        """Determine gain and bias by shifting and scaling the lines."""
        max_rates = np.array(max_rates, dtype=float, copy=False, ndmin=1)
        intercepts = np.array(intercepts, dtype=float, copy=False, ndmin=1)
        gain = max_rates / (1 - intercepts)
        bias = -intercepts * gain
        return gain, bias

    def max_rates_intercepts(self, gain, bias):
        """Compute the inverse of gain_bias."""
        intercepts = -bias / gain
        max_rates = gain * (1 - intercepts)
        return max_rates, intercepts

    def step(self, dt, J, output):
        """Implement the rectification nonlinearity."""
        output[...] = self.amplitude * np.maximum(0.0, J)


class SpikingRectifiedLinear(RectifiedLinear):
    """A rectified integrate and fire neuron model.

    Each neuron is modeled as a rectified line. That is, the neuron's activity
    scales linearly with current, unless the current is less than zero, at
    which point the neural activity will stay at zero. This is a spiking
    version of the RectifiedLinear neuron model.

    Parameters
    ----------
    amplitude : float
        Scaling factor on the neuron output. Corresponds to the relative
        amplitude of the output spikes of the neuron.
    initial_state : {str: Distribution or array_like}
        Mapping from state variables names to their desired initial value.
        These values will override the defaults set in the class's state attribute.
    """

    state = {"voltage": Uniform(low=0, high=1)}
    spiking = True

    def rates(self, x, gain, bias):
        """Use RectifiedLinear to determine rates."""

        J = self.current(x, gain, bias)
        out = np.zeros_like(J)
        RectifiedLinear.step(self, dt=1.0, J=J, output=out)
        return out

    def step(self, dt, J, output, voltage):
        """Implement the integrate and fire nonlinearity."""

        voltage += np.maximum(J, 0) * dt
        n_spikes = np.floor(voltage)
        output[:] = (self.amplitude / dt) * n_spikes
        voltage -= n_spikes


class Sigmoid(NeuronType):
    """A non-spiking neuron model whose response curve is a sigmoid.

    Since the tuning curves are strictly positive, the ``intercepts``
    correspond to the inflection point of each sigmoid. That is,
    ``f(intercept) = 0.5`` where ``f`` is the pure sigmoid function.

    Parameters
    ----------
    tau_ref : float
        The neuron refractory period, in seconds. The maximum firing rate of the
        neurons is ``1 / tau_ref``. Must be positive (i.e. ``tau_ref > 0``).
    initial_state : {str: Distribution or array_like}
        Mapping from state variables names to their desired initial value.
        These values will override the defaults set in the class's state attribute.
    """

    negative = False

    tau_ref = NumberParam("tau_ref", low=0, low_open=True)

    def __init__(self, tau_ref=0.0025, initial_state=None):
        super().__init__(initial_state)
        self.tau_ref = tau_ref

    def gain_bias(self, max_rates, intercepts):
        """Analytically determine gain, bias."""
        max_rates = np.array(max_rates, dtype=float, copy=False, ndmin=1)
        intercepts = np.array(intercepts, dtype=float, copy=False, ndmin=1)

        inv_tau_ref = 1.0 / self.tau_ref
        if not np.all(max_rates < inv_tau_ref):
            raise ValidationError(
                "Max rates must be below the inverse "
                f"refractory period ({inv_tau_ref:0.3f})",
                attr="max_rates",
                obj=self,
            )

        inverse = -np.log(inv_tau_ref / max_rates - 1.0)
        gain = inverse / (1.0 - intercepts)
        bias = inverse - gain
        return gain, bias

    def max_rates_intercepts(self, gain, bias):
        """Compute the inverse of gain_bias."""
        inverse = gain + bias
        intercepts = 1 - inverse / gain
        max_rates = (1.0 / self.tau_ref) / (1 + np.exp(-inverse))
        return max_rates, intercepts

    def step(self, dt, J, output):
        """Implement the sigmoid nonlinearity."""
        output[...] = (1.0 / self.tau_ref) / (1 + np.exp(-J))


class Tanh(NeuronType):
    """A non-spiking neuron model whose response curve is a hyperbolic tangent.

    Parameters
    ----------
    tau_ref : float
        The neuron refractory period, in seconds. The maximum firing rate of the
        neurons is ``1 / tau_ref``. Must be positive (i.e. ``tau_ref > 0``).
    initial_state : {str: Distribution or array_like}
        Mapping from state variables names to their desired initial value.
        These values will override the defaults set in the class's state attribute.
    """

    tau_ref = NumberParam("tau_ref", low=0, low_open=True)

    def __init__(self, tau_ref=0.0025, initial_state=None):
        super().__init__(initial_state)
        self.tau_ref = tau_ref

    def gain_bias(self, max_rates, intercepts):
        """Analytically determine gain, bias."""
        max_rates = np.array(max_rates, dtype=float, copy=False, ndmin=1)
        intercepts = np.array(intercepts, dtype=float, copy=False, ndmin=1)

        inv_tau_ref = 1.0 / self.tau_ref
        if not np.all(max_rates < inv_tau_ref):
            raise ValidationError(
                "Max rates must be below the inverse "
                f"refractory period ({inv_tau_ref:0.3f})",
                attr="max_rates",
                obj=self,
            )

        inverse = np.arctanh(max_rates * self.tau_ref)
        gain = inverse / (1.0 - intercepts)
        bias = -gain * intercepts
        return gain, bias

    def max_rates_intercepts(self, gain, bias):
        """Compute the inverse of gain_bias."""
        intercepts = -bias / gain
        max_rates = (1.0 / self.tau_ref) * np.tanh(gain + bias)
        return max_rates, intercepts

    def step(self, dt, J, output):
        """Implement the tanh nonlinearity."""
        output[...] = (1.0 / self.tau_ref) * np.tanh(J)


class LIFRate(NeuronType):
    """Non-spiking version of the leaky integrate-and-fire (LIF) neuron model.

    Parameters
    ----------
    tau_rc : float
        Membrane RC time constant, in seconds. Affects how quickly the membrane
        voltage decays to zero in the absence of input (larger = slower decay).
    tau_ref : float
        Absolute refractory period, in seconds. This is how long the
        membrane voltage is held at zero after a spike.
    amplitude : float
        Scaling factor on the neuron output. Corresponds to the relative
        amplitude of the output spikes of the neuron.
    initial_state : {str: Distribution or array_like}
        Mapping from state variables names to their desired initial value.
        These values will override the defaults set in the class's state attribute.
    """

    negative = False

    tau_rc = NumberParam("tau_rc", low=0, low_open=True)
    tau_ref = NumberParam("tau_ref", low=0)
    amplitude = NumberParam("amplitude", low=0, low_open=True)

    def __init__(self, tau_rc=0.02, tau_ref=0.002, amplitude=1, initial_state=None):
        super().__init__(initial_state)
        self.tau_rc = tau_rc
        self.tau_ref = tau_ref
        self.amplitude = amplitude

    def gain_bias(self, max_rates, intercepts):
        """Analytically determine gain, bias."""
        max_rates = np.array(max_rates, dtype=float, copy=False, ndmin=1)
        intercepts = np.array(intercepts, dtype=float, copy=False, ndmin=1)

        inv_tau_ref = 1.0 / self.tau_ref if self.tau_ref > 0 else np.inf
        if not np.all(max_rates < inv_tau_ref):
            raise ValidationError(
                "Max rates must be below the inverse "
                f"refractory period ({inv_tau_ref:0.3f})",
                attr="max_rates",
                obj=self,
            )

        x = 1.0 / (1 - np.exp((self.tau_ref - (1.0 / max_rates)) / self.tau_rc))
        gain = (1 - x) / (intercepts - 1.0)
        bias = 1 - gain * intercepts
        return gain, bias

    def max_rates_intercepts(self, gain, bias):
        """Compute the inverse of gain_bias."""
        intercepts = (1 - bias) / gain
        max_rates = 1.0 / (
            self.tau_ref - self.tau_rc * np.log1p(1.0 / (gain * (intercepts - 1) - 1))
        )
        if not np.all(np.isfinite(max_rates)):
            warnings.warn(
                "Non-finite values detected in `max_rates`; this "
                "probably means that `gain` was too small."
            )
        return max_rates, intercepts

    def rates(self, x, gain, bias):
        """Always use LIFRate to determine rates."""
        J = self.current(x, gain, bias)
        out = np.zeros_like(J)
        # Use LIFRate's step explicitly to ensure rate approximation
        LIFRate.step(self, dt=1, J=J, output=out)
        return out

    def step(self, dt, J, output):
        """Implement the LIFRate nonlinearity."""
        j = J - 1
        output[:] = 0  # faster than output[j <= 0] = 0
        output[j > 0] = self.amplitude / (
            self.tau_ref + self.tau_rc * np.log1p(1.0 / j[j > 0])
        )
        # the above line is designed to throw an error if any j is nan
        # (nan > 0 -> error), and not pass x < -1 to log1p


class LIF(LIFRate):
    """Spiking version of the leaky integrate-and-fire (LIF) neuron model.

    Parameters
    ----------
    tau_rc : float
        Membrane RC time constant, in seconds. Affects how quickly the membrane
        voltage decays to zero in the absence of input (larger = slower decay).
    tau_ref : float
        Absolute refractory period, in seconds. This is how long the
        membrane voltage is held at zero after a spike.
    min_voltage : float
        Minimum value for the membrane voltage. If ``-np.inf``, the voltage
        is never clipped.
    amplitude : float
        Scaling factor on the neuron output. Corresponds to the relative
        amplitude of the output spikes of the neuron.
    initial_state : {str: Distribution or array_like}
        Mapping from state variables names to their desired initial value.
        These values will override the defaults set in the class's state attribute.
    """

    state = {
        "voltage": Uniform(low=0, high=1),
        "refractory_time": Choice([0]),
    }
    spiking = True

    min_voltage = NumberParam("min_voltage", high=0)

    def __init__(
        self, tau_rc=0.02, tau_ref=0.002, min_voltage=0, amplitude=1, initial_state=None
    ):
        super().__init__(
            tau_rc=tau_rc,
            tau_ref=tau_ref,
            amplitude=amplitude,
            initial_state=initial_state,
        )
        self.min_voltage = min_voltage

    def step(self, dt, J, output, voltage, refractory_time):
        # look these up once to avoid repeated parameter accesses
        tau_rc = self.tau_rc
        min_voltage = self.min_voltage

        # reduce all refractory times by dt
        refractory_time -= dt

        # compute effective dt for each neuron, based on remaining time.
        # note that refractory times that have completed midway into this
        # timestep will be given a partial timestep, and moreover these will
        # be subtracted to zero at the next timestep (or reset by a spike)
        delta_t = clip((dt - refractory_time), 0, dt)

        # update voltage using discretized lowpass filter
        # since v(t) = v(0) + (J - v(0))*(1 - exp(-t/tau)) assuming
        # J is constant over the interval [t, t + dt)
        voltage -= (J - voltage) * np.expm1(-delta_t / tau_rc)

        # determine which neurons spiked (set them to 1/dt, else 0)
        spiked_mask = voltage > 1
        output[:] = spiked_mask * (self.amplitude / dt)

        # set v(0) = 1 and solve for t to compute the spike time
        t_spike = dt + tau_rc * np.log1p(
            -(voltage[spiked_mask] - 1) / (J[spiked_mask] - 1)
        )

        # set spiked voltages to zero, refractory times to tau_ref, and
        # rectify negative voltages to a floor of min_voltage
        voltage[voltage < min_voltage] = min_voltage
        voltage[spiked_mask] = 0
        refractory_time[spiked_mask] = self.tau_ref + t_spike


class AdaptiveLIFRate(LIFRate):
    """Adaptive non-spiking version of the LIF neuron model.

    Works as the LIF model, except with adapation state ``n``, which is
    subtracted from the input current. Its dynamics are::

        tau_n dn/dt = -n

    where ``n`` is incremented by ``inc_n`` when the neuron spikes.

    Parameters
    ----------
    tau_n : float
        Adaptation time constant. Affects how quickly the adaptation state
        decays to zero in the absence of spikes (larger = slower decay).
    inc_n : float
        Adaptation increment. How much the adaptation state is increased after
        each spike.
    tau_rc : float
        Membrane RC time constant, in seconds. Affects how quickly the membrane
        voltage decays to zero in the absence of input (larger = slower decay).
    tau_ref : float
        Absolute refractory period, in seconds. This is how long the
        membrane voltage is held at zero after a spike.
    amplitude : float
        Scaling factor on the neuron output. Corresponds to the relative
        amplitude of the output spikes of the neuron.
    initial_state : {str: Distribution or array_like}
        Mapping from state variables names to their desired initial value.
        These values will override the defaults set in the class's state attribute.

    References
    ----------
    .. [1] Camera, Giancarlo La, et al. "Minimal models of adapted neuronal
       response to in Vivo-Like input currents." Neural computation
       16.10 (2004): 2101-2124.
    """

    state = {"adaptation": Choice([0])}

    tau_n = NumberParam("tau_n", low=0, low_open=True)
    inc_n = NumberParam("inc_n", low=0)

    def __init__(
        self,
        tau_n=1,
        inc_n=0.01,
        tau_rc=0.02,
        tau_ref=0.002,
        amplitude=1,
        initial_state=None,
    ):
        super().__init__(
            tau_rc=tau_rc,
            tau_ref=tau_ref,
            amplitude=amplitude,
            initial_state=initial_state,
        )
        self.tau_n = tau_n
        self.inc_n = inc_n

    def step(self, dt, J, output, adaptation):
        """Implement the AdaptiveLIFRate nonlinearity."""
        n = adaptation
        super().step(dt, J - n, output)
        n += (dt / self.tau_n) * (self.inc_n * output - n)


class AdaptiveLIF(LIF):
    """Adaptive spiking version of the LIF neuron model.

    Works as the LIF model, except with adapation state ``n``, which is
    subtracted from the input current. Its dynamics are::

        tau_n dn/dt = -n

    where ``n`` is incremented by ``inc_n`` when the neuron spikes.

    Parameters
    ----------
    tau_n : float
        Adaptation time constant. Affects how quickly the adaptation state
        decays to zero in the absence of spikes (larger = slower decay).
    inc_n : float
        Adaptation increment. How much the adaptation state is increased after
        each spike.
    tau_rc : float
        Membrane RC time constant, in seconds. Affects how quickly the membrane
        voltage decays to zero in the absence of input (larger = slower decay).
    tau_ref : float
        Absolute refractory period, in seconds. This is how long the
        membrane voltage is held at zero after a spike.
    min_voltage : float
        Minimum value for the membrane voltage. If ``-np.inf``, the voltage
        is never clipped.
    amplitude : float
        Scaling factor on the neuron output. Corresponds to the relative
        amplitude of the output spikes of the neuron.
    initial_state : {str: Distribution or array_like}
        Mapping from state variables names to their desired initial value.
        These values will override the defaults set in the class's state attribute.

    References
    ----------
    .. [1] Camera, Giancarlo La, et al. "Minimal models of adapted neuronal
       response to in Vivo-Like input currents." Neural computation
       16.10 (2004): 2101-2124.
    """

    state = {
        "voltage": Uniform(low=0, high=1),
        "refractory_time": Choice([0]),
        "adaptation": Choice([0]),
    }
    spiking = True

    tau_n = NumberParam("tau_n", low=0, low_open=True)
    inc_n = NumberParam("inc_n", low=0)

    def __init__(
        self,
        tau_n=1,
        inc_n=0.01,
        tau_rc=0.02,
        tau_ref=0.002,
        min_voltage=0,
        amplitude=1,
        initial_state=None,
    ):
        super().__init__(
            tau_rc=tau_rc,
            tau_ref=tau_ref,
            min_voltage=min_voltage,
            amplitude=amplitude,
            initial_state=initial_state,
        )
        self.tau_n = tau_n
        self.inc_n = inc_n

    def step(self, dt, J, output, voltage, refractory_time, adaptation):
        """Implement the AdaptiveLIF nonlinearity."""
        n = adaptation
        super().step(dt, J - n, output, voltage, refractory_time)
        n += (dt / self.tau_n) * (self.inc_n * output - n)


class Izhikevich(NeuronType):
    """Izhikevich neuron model.

    This implementation is based on the original paper [1]_;
    however, we rename some variables for clarity.
    What was originally 'v' we term 'voltage', which represents the membrane
    potential of each neuron. What was originally 'u' we term 'recovery',
    which represents membrane recovery, "which accounts for the activation
    of K+ ionic currents and inactivation of Na+ ionic currents."
    The 'a', 'b', 'c', and 'd' parameters are also renamed
    (see the parameters below).

    We use default values that correspond to regular spiking ('RS') neurons.
    For other classes of neurons, set the parameters as follows.

    * Intrinsically bursting (IB): ``reset_voltage=-55, reset_recovery=4``
    * Chattering (CH): ``reset_voltage=-50, reset_recovery=2``
    * Fast spiking (FS): ``tau_recovery=0.1``
    * Low-threshold spiking (LTS): ``coupling=0.25``
    * Resonator (RZ): ``tau_recovery=0.1, coupling=0.26``

    Parameters
    ----------
    tau_recovery : float, optional
        (Originally 'a') Time scale of the recovery variable.
    coupling : float, optional
        (Originally 'b') How sensitive recovery is to subthreshold
        fluctuations of voltage.
    reset_voltage : float, optional
        (Originally 'c') The voltage to reset to after a spike, in millivolts.
    reset_recovery : float, optional
        (Originally 'd') The recovery value to reset to after a spike.
    initial_state : {str: Distribution or array_like}
        Mapping from state variables names to their desired initial value.
        These values will override the defaults set in the class's state attribute.

    References
    ----------
    .. [1] E. M. Izhikevich, "Simple model of spiking neurons."
       IEEE Transactions on Neural Networks, vol. 14, no. 6, pp. 1569-1572.
       (http://www.izhikevich.org/publications/spikes.pdf)
    """

    state = {
        "voltage": Uniform(low=0, high=1),
        "recovery": Choice([0]),
    }
    negative = False
    spiking = True

    tau_recovery = NumberParam("tau_recovery", low=0, low_open=True)
    coupling = NumberParam("coupling", low=0)
    reset_voltage = NumberParam("reset_voltage")
    reset_recovery = NumberParam("reset_recovery")

    def __init__(
        self,
        tau_recovery=0.02,
        coupling=0.2,
        reset_voltage=-65.0,
        reset_recovery=8.0,
        initial_state=None,
    ):
        super().__init__(initial_state)
        self.tau_recovery = tau_recovery
        self.coupling = coupling
        self.reset_voltage = reset_voltage
        self.reset_recovery = reset_recovery

    def rates(self, x, gain, bias):
        """Estimates steady-state firing rate given gain and bias."""
        J = self.current(x, gain, bias)
        return settled_firingrate(
            self.step,
            J,
            state={
                "output": np.zeros_like(J),
                "voltage": np.zeros_like(J),
                "recovery": np.zeros_like(J),
            },
            settle_time=0.001,
            sim_time=1.0,
        )

    def step(self, dt, J, output, voltage, recovery):
        """Implement the Izhikevich nonlinearity."""
        # Numerical instability occurs for very low inputs.
        # We'll clip them be greater than some value that was chosen by
        # looking at the simulations for many parameter sets.
        # A more principled minimum value would be better.
        J = np.maximum(-30.0, J)

        dV = (0.04 * voltage ** 2 + 5 * voltage + 140 - recovery + J) * 1000
        voltage[:] += dV * dt

        # We check for spikes and reset the voltage here rather than after,
        # which differs from the original implementation by Izhikevich.
        # However, calculating recovery for voltage values greater than
        # threshold can cause the system to blow up, which we want
        # to avoid at all costs.
        output[:] = (voltage >= 30) / dt
        voltage[output > 0] = self.reset_voltage

        dU = (self.tau_recovery * (self.coupling * voltage - recovery)) * 1000
        recovery[:] += dU * dt
        recovery[output > 0] = recovery[output > 0] + self.reset_recovery


class RatesToSpikesNeuronType(NeuronType):
    """Base class for neuron types that turn rate types into spiking ones."""

    base_type = NeuronTypeParam("base_type")
    amplitude = NumberParam("amplitude", low=0, low_open=True)
    spiking = True

    def __init__(self, base_type, amplitude=1.0, initial_state=None):
        super().__init__(initial_state)

        self.base_type = base_type
        self.amplitude = amplitude
        self.negative = base_type.negative

        if base_type.spiking:
            warnings.warn(
                f"'base_type' is type '{type(base_type).__name__}', which is a spiking "
                "neuron type. We recommend using the non-spiking equivalent type, "
                "if one exists."
            )

        for s in self.state:
            if s in self.base_type.state:
                raise ValidationError(
                    f"{self} and {self.base_type} have an overlapping "
                    f"state variable ({s})",
                    attr="state",
                    obj=self,
                )

    def gain_bias(self, max_rates, intercepts):
        return self.base_type.gain_bias(max_rates, intercepts)

    def max_rates_intercepts(self, gain, bias):
        return self.base_type.max_rates_intercepts(gain, bias)

    def rates(self, x, gain, bias):
        return self.base_type.rates(x, gain, bias)

    def step(self, dt, J, output, **state):
        raise NotImplementedError("Subclasses must implement step")

    @property
    def probeable(self):
        return ("output", "rate_out") + tuple(self.state) + tuple(self.base_type.state)


class RegularSpiking(RatesToSpikesNeuronType):
    """Turn a rate neuron type into a spiking one with regular inter-spike intervals.

    Spikes at regular intervals based on the rates of the base neuron type. [1]_

    Parameters
    ----------
    base_type : NeuronType
        A rate-based neuron type to convert to a regularly spiking neuron.
    amplitude : float
        Scaling factor on the neuron output. Corresponds to the relative
        amplitude of the output spikes of the neuron.
    initial_state : {str: Distribution or array_like}
        Mapping from state variables names to their desired initial value.
        These values will override the defaults set in the class's state attribute.

    References
    ----------
    .. [1] Voelker, A. R., Rasmussen, D., & Eliasmith, C. (2020). A Spike in
       Performance: Training Hybrid-Spiking Neural Networks with Quantized Activation
       Functions. arXiv preprint arXiv:2002.03553.
       (https://export.arxiv.org/abs/2002.03553)
    """

    state = {"voltage": Uniform(low=0, high=1)}

    def step(self, dt, J, output, voltage):
        # Note: J is the desired output rate, not the input current
        voltage += dt * J
        n_spikes = np.floor(voltage)
        output[...] = (self.amplitude / dt) * n_spikes
        voltage -= n_spikes


class StochasticSpiking(RatesToSpikesNeuronType):
    """Turn a rate neuron type into a spiking one using stochastic rounding.

    The expected number of spikes per timestep ``e = dt * r`` is determined by the
    base type firing rate ``r`` and the timestep ``dt``. Given the fractional part ``f``
    and integer part ``q`` of ``e``, the number of generated spikes is ``q`` with
    probability ``1 - f`` and ``q + 1`` with probability ``f``. For ``e`` much less than
    one, this is very similar to Poisson statistics.

    Parameters
    ----------
    base_type : NeuronType
        A rate-based neuron type to convert to a stochastic spiking neuron.
    amplitude : float
        Scaling factor on the neuron output. Corresponds to the relative
        amplitude of the output spikes of the neuron.
    initial_state : {str: Distribution or array_like}
        Mapping from state variables names to their desired initial value.
        These values will override the defaults set in the class's state attribute.
    """

    def make_state(self, n_neurons, rng=np.random, dtype=None):
        state = super().make_state(n_neurons, rng=rng, dtype=dtype)
        state["rng"] = rng
        return state

    def step(self, dt, J, output, rng, **base_state):
        # Note: J is the desired output rate, not the input current
        if self.negative:
            frac, n_spikes = np.modf(dt * np.abs(J))
        else:
            frac, n_spikes = np.modf(dt * J)

        n_spikes += rng.random_sample(size=frac.shape) < frac

        if self.negative:
            output[...] = (self.amplitude / dt) * n_spikes * np.sign(J)
        else:
            output[...] = (self.amplitude / dt) * n_spikes


class PoissonSpiking(RatesToSpikesNeuronType):
    """Turn a rate neuron type into a spiking one with Poisson spiking statistics.

    Spikes with Poisson probability based on the rates of the base neuron type.

    Parameters
    ----------
    base_type : NeuronType
        A rate-based neuron type to convert to a Poisson spiking neuron.
    amplitude : float
        Scaling factor on the neuron output. Corresponds to the relative
        amplitude of the output spikes of the neuron.
    initial_state : {str: Distribution or array_like}
        Mapping from state variables names to their desired initial value.
        These values will override the defaults set in the class's state attribute.
    """

    def make_state(self, n_neurons, rng=np.random, dtype=None):
        state = super().make_state(n_neurons, rng=rng, dtype=dtype)
        state["rng"] = rng
        return state

    def step(self, dt, J, output, rng, **base_state):
        # Note: J is the desired output rate, not the input current
        if self.negative:
            output[...] = (
                (self.amplitude / dt)
                * rng.poisson(np.abs(J) * dt, output.size)
                * np.sign(J)
            )
        else:
            output[...] = (self.amplitude / dt) * rng.poisson(J * dt, output.size)



from collections import Counter

import pytest

import nengo
from nengo.exceptions import ConfigError, NetworkContextError, ReadonlyError
from nengo.utils.testing import ThreadedAssertion


def test_basic_context():
    # We must give the two Networks different labels because object comparison
    # is done using identifiers that stem from the top-level Network label.
    model1 = nengo.Network(label="test1")
    model2 = nengo.Network(label="test2")

    with model1:
        e = nengo.Ensemble(1, dimensions=1)
        n = nengo.Node([0])
        assert e in model1.ensembles
        assert n in model1.nodes

        con = nengo.Network()
        with con:
            e2 = nengo.Ensemble(1, dimensions=1)
        assert e2 in con.ensembles
        assert e2 not in model1.ensembles

        e3 = nengo.Ensemble(1, dimensions=1)
        assert e3 in model1.ensembles

    with model2:
        e4 = nengo.Ensemble(1, dimensions=1)
        assert e4 not in model1.ensembles
        assert e4 in model2.ensembles


def test_nested_context():
    model = nengo.Network()
    with model:
        con1 = nengo.Network()
        con2 = nengo.Network()
        con3 = nengo.Network()

        with con1:
            e1 = nengo.Ensemble(1, dimensions=1)
            assert e1 in con1.ensembles

            with con2:
                e2 = nengo.Ensemble(1, dimensions=1)
                assert e2 in con2.ensembles
                assert e2 not in con1.ensembles

                with con3:
                    e3 = nengo.Ensemble(1, dimensions=1)
                    assert e3 in con3.ensembles
                    assert e3 not in con2.ensembles and e3 not in con1.ensembles

                e4 = nengo.Ensemble(1, dimensions=1)
                assert e4 in con2.ensembles
                assert e4 not in con3.ensembles

            e5 = nengo.Ensemble(1, dimensions=1)
            assert e5 in con1.ensembles

        e6 = nengo.Ensemble(1, dimensions=1)
        assert e6 not in con1.ensembles


def test_context_errors(request):
    """Most of these errors are in ``Network.add``; one is in ``Network.__exit__``"""

    def clear_contexts():  # clear anything we've added to the contexts
        nengo.Network.context.clear()
        nengo.Config.context.clear()

    request.addfinalizer(clear_contexts)

    def add_something():
        nengo.Ensemble(1, dimensions=1)

    # Error if adding before Network creation
    with pytest.raises(NetworkContextError, match="must either be created inside a"):
        add_something()

    model = nengo.Network()
    # Error if adding before a `with network` block
    with pytest.raises(NetworkContextError, match="must either be created inside a"):
        add_something()

    # Error if adding after a `with network` block
    with model:
        add_something()
    with pytest.raises(NetworkContextError, match="must either be created inside a"):
        add_something()

    # Error if adding a bad object type
    with model:
        with pytest.raises(NetworkContextError, match="Objects of type.*added to net"):
            model.add(object())

    # Okay if add_to_container=False
    nengo.Ensemble(1, dimensions=1, add_to_container=False)
    nengo.Node(output=[0], add_to_container=False)

    # Errors if context is in a bad state (one in `Network.__exit__`)
    with pytest.raises(NetworkContextError, match="bad state; was expecting current"):
        with nengo.Network():
            nengo.Network.context.append("bad")
            with pytest.raises(NetworkContextError, match="context (bad) is not"):
                nengo.Ensemble(10, 1)

    clear_contexts()

    # Error if config is in a bad state
    with pytest.raises(ConfigError, match="bad state; was expecting current"):
        with nengo.Network():
            nengo.Config.context.append(nengo.Config(nengo.Ensemble))

    clear_contexts()

    # Error if context is empty when exiting
    with pytest.raises(NetworkContextError, match="bad state; was empty when exiting"):
        with nengo.Network():
            nengo.Network.context.clear()


def test_context_is_threadsafe():
    class CheckIndependence(ThreadedAssertion):
        def init_thread(self, worker):
            setattr(worker, "model", nengo.Network())
            worker.model.__enter__()

        def assert_thread(self, worker):
            assert list(nengo.Network.context) == [worker.model]

        def finish_thread(self, worker):
            worker.model.__exit__(*worker.exc_info)

    CheckIndependence(n_threads=2).run()


def test_get_objects():
    model = nengo.Network()
    with model:
        ens1 = nengo.Ensemble(10, 1)
        node1 = nengo.Node([0])
        conn1 = nengo.Connection(node1, ens1)
        pr1 = nengo.Probe(ens1)

        subnet = nengo.Network()

        with subnet:
            ens2 = nengo.Ensemble(10, 1)
            node2 = nengo.Node([0])
            conn2 = nengo.Connection(node2, ens2)
            pr2 = nengo.Probe(ens2)
            subnet2 = nengo.Network()

            with subnet2:
                ens3 = nengo.Ensemble(10, 1)

    all_objects = [
        ens1,
        pr1,
        node1,
        conn1,
        subnet,
        ens2,
        node2,
        conn2,
        pr2,
        ens3,
        subnet2,
    ]

    # Test that the lists contain the same elements, but order doesn't matter.
    # Counter is like a set, but also keeps track of the number of objects.
    assert Counter(all_objects) == Counter(model.all_objects)
    assert Counter([ens1, ens2, ens3]) == Counter(model.all_ensembles)
    assert Counter([node1, node2]) == Counter(model.all_nodes)
    assert Counter([conn1, conn2]) == Counter(model.all_connections)
    assert Counter([pr1, pr2]) == Counter(model.all_probes)
    assert Counter([subnet, subnet2]) == Counter(model.all_networks)
    # Make sure it works a second time
    assert Counter([ens1, ens2, ens3]) == Counter(model.all_ensembles)


def test_raises_exception_on_incompatiple_type_arguments():
    with pytest.raises(ValueError):
        nengo.Network(label=1)
    with pytest.raises(ValueError):
        nengo.Network(seed="label")


def test_n_neurons():
    with nengo.Network() as net:
        nengo.Ensemble(10, 1)
        assert net.n_neurons == 10
        with nengo.Network() as subnet:
            nengo.Ensemble(30, 1)
            assert subnet.n_neurons == 30
        assert net.n_neurons == 40


def test_readonly_attributes():
    def test_attr(net, attr):
        with pytest.raises(ReadonlyError, match=f"Network.{attr}: {attr} is read-only"):
            setattr(net, attr, [])

    with nengo.Network() as net:
        for attr in (
            "objects",
            "ensembles",
            "nodes",
            "networks",
            "connections",
            "probes",
            "config",
        ):
            test_attr(net, attr)


def test_network_contains():
    out_ens = nengo.Ensemble(10, 1, add_to_container=False)
    with nengo.Network() as net:
        ens = nengo.Ensemble(10, 1)
        assert ens in net
        assert out_ens not in net
        assert "str" not in net



import numpy as np

from nengo.builder.builder import Builder
from nengo.builder.operator import Operator
from nengo.builder.signal import Signal
from nengo.processes import Process
from nengo.rc import rc


class SimProcess(Operator):
    """Simulate a process.

    Parameters
    ----------
    process : Process
        The `.Process` to simulate.
    input : Signal or None
        Input to the process, or None if no input.
    output : Signal or None
        Output from the process, or None if no output.
    t : Signal
        The signal associated with the time (a float, in seconds).
    mode : str, optional
        Denotes what type of update this operator performs.
        Must be one of ``'update'``, ``'inc'`` or ``'set'``.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Attributes
    ----------
    input : Signal or None
        Input to the process, or None if no input.
    mode : str
        Denotes what type of update this operator performs.
    output : Signal or None
        Output from the process, or None if no output.
    process : Process
        The `.Process` to simulate.
    t : Signal
        The signal associated with the time (a float, in seconds).
    tag : str or None
        A label associated with the operator, for debugging purposes.

    Notes
    -----
    1. sets ``[output] if output is not None and mode=='set' else []``
    2. incs ``[output] if output is not None and mode=='inc' else []``
    3. reads ``[t, input] if input is not None else [t]``
    4. updates ``[output] if output is not None and mode=='update' else []``
    """

    def __init__(self, process, input, output, t, mode="set", state=None, tag=None):
        super().__init__(tag=tag)
        self.process = process
        self.mode = mode

        assert output is not None

        self.reads = [t, input] if input is not None else [t]
        self.sets = []
        self.incs = []
        self.updates = []
        if mode == "update":
            self.updates.extend([output])
        elif mode == "inc":
            self.incs.extend([output])
        elif mode == "set":
            self.sets.extend([output])
        else:
            raise ValueError(f"Unrecognized mode '{mode}'")

        self.state_idxs = {}
        if state is not None:
            for name, sig in state.items():
                # The signals actually stored in `self.updates` can be modified by the
                # optimizer. To allow this possibility, we store the index of the
                # signal in the updates list instead of storing the signal itself.
                self.state_idxs[name] = len(self.updates)
                self.updates.append(sig)

    @property
    def input(self):
        return None if len(self.reads) == 1 else self.reads[1]

    @property
    def output(self):
        if self.mode == "update":
            return self.updates[0]
        if self.mode == "inc":
            return self.incs[0]
        assert self.mode == "set"
        return self.sets[0]

    @property
    def state(self):
        return {key: self.updates[idx] for key, idx in self.state_idxs.items()}

    @property
    def t(self):
        return self.reads[0]

    @property
    def _descstr(self):
        return f"{self.process}, {self.input} -> {self.output}"

    def make_step(self, signals, dt, rng):
        t = signals[self.t]
        input = signals[self.input] if self.input is not None else None
        output = signals[self.output]
        shape_in = input.shape if input is not None else (0,)
        shape_out = output.shape
        rng = self.process.get_rng(rng)
        state = {name: signals[sig] for name, sig in self.state.items()}
        step_f = self.process.make_step(shape_in, shape_out, dt, rng, state)

        if self.mode == "inc" and input is None:

            def step_simprocess():
                output[...] += step_f(t.item())

        elif self.mode == "inc":

            def step_simprocess():
                output[...] += step_f(t.item(), np.copy(input))

        elif input is None:

            def step_simprocess():
                output[...] = step_f(t.item())

        else:
            assert self.mode != "inc" and input is not None

            def step_simprocess():
                output[...] = step_f(t.item(), np.copy(input))

        return step_simprocess


@Builder.register(Process)
def build_process(model, process, sig_in=None, sig_out=None, mode="set"):
    """Builds a `.Process` object into a model.

    Parameters
    ----------
    model : Model
        The model to build into.
    process : Process
        Process to build.
    sig_in : Signal, optional
        The input signal, or None if no input signal.
    sig_out : Signal, optional
        The output signal, or None if no output signal.
    mode : "set" or "inc" or "update", optional
        The ``mode`` of the built `.SimProcess`.

    Notes
    -----
    Does not modify ``model.params[]`` and can therefore be called
    more than once with the same `.Process` instance.
    """
    if sig_out is None:
        sig_out = Signal(shape=sig_in.shape, name=f"{sig_in.name}.{process}")

    shape_in = sig_in.shape if sig_in is not None else (0,)
    shape_out = sig_out.shape if sig_out is not None else (0,)
    dtype = (
        sig_out.dtype
        if sig_out is not None
        else sig_in.dtype
        if sig_in is not None
        else rc.float_dtype
    )
    state_init = process.make_state(shape_in, shape_out, model.dt, dtype=dtype)
    state = {}
    for name, value in state_init.items():
        state[name] = Signal(value)
        model.sig[process]["_state_" + name] = state[name]

    model.add_op(
        SimProcess(
            process,
            sig_in,
            sig_out,
            model.time,
            mode=mode,
            state=state,
            tag=str(process),
        )
    )

    return sig_out



from . import (
    builder,
    cache,
    connection,
    ensemble,
    filter_design,
    functions,
    graphs,
    ipython,
    magic,
    nco,
    network,
    numpy,
    paths,
    simulator,
    stdlib,
)



import numpy as np
import pytest

from nengo.builder.transforms import multiply
from nengo.exceptions import BuildError


def test_multiply():
    sca = np.array(4)
    vec = np.array([2, 3])
    mat = np.array([[1, 2], [3, 4]])

    assert np.array_equal(multiply(sca, sca), sca * sca)
    assert np.array_equal(multiply(sca, vec), sca * vec)
    assert np.array_equal(multiply(vec, sca), sca * vec)
    assert np.array_equal(multiply(sca, mat), sca * mat)
    assert np.array_equal(multiply(mat, sca), sca * mat)

    assert np.array_equal(multiply(vec, vec), vec * vec)
    assert np.array_equal(multiply(vec, mat), np.diag(vec).dot(mat))
    assert np.array_equal(multiply(mat, vec), mat.dot(np.diag(vec)))
    assert np.array_equal(multiply(mat, mat), mat.dot(mat))

    with pytest.raises(BuildError):
        ary3 = np.ones((2, 2, 2))
        multiply(ary3, mat)



import pickle
from fnmatch import fnmatch

import pytest

import nengo
import nengo.synapses
from nengo.config import ClassParams, InstanceParams, SupportDefaultsMixin
from nengo.exceptions import ConfigError, ReadonlyError
from nengo.params import Default, Parameter, Unconfigurable
from nengo.utils.testing import ThreadedAssertion


def test_config_basic():
    model = nengo.Network()
    model.config[nengo.Ensemble].set_param("something", Parameter("something", None))
    model.config[nengo.Ensemble].set_param("other", Parameter("other", default=0))
    model.config[nengo.Connection].set_param(
        "something_else", Parameter("something_else", None)
    )

    with pytest.raises(ConfigError, match="'fails' is not a parameter"):
        model.config[nengo.Ensemble].set_param("fails", 1.0)

    with pytest.raises(ConfigError, match="'bias' is already a parameter in"):
        model.config[nengo.Ensemble].set_param("bias", Parameter("bias", None))

    with model:
        a = nengo.Ensemble(50, dimensions=1)
        b = nengo.Ensemble(90, dimensions=1)
        a2b = nengo.Connection(a, b, synapse=0.01)

    with pytest.raises(ConfigError, match="Cannot set parameters on an instance"):
        model.config[a].set_param("thing", Parameter("thing", None))

    with pytest.raises(ConfigError, match="Cannot get parameters on an instance"):
        model.config[a].get_param("bias")

    assert model.config[a].something is None
    assert model.config[b].something is None
    assert model.config[a].other == 0
    assert model.config[b].other == 0
    assert model.config[a2b].something_else is None

    model.config[a].something = "hello"
    assert model.config[a].something == "hello"
    model.config[a].something = "world"
    assert model.config[a].something == "world"
    del model.config[a].something
    assert model.config[a].something is None

    with pytest.raises(AttributeError):
        print(model.config[a].something_else)
    with pytest.raises(AttributeError):
        print(model.config[a2b].something)
    with pytest.raises(AttributeError):
        model.config[a].something_else = 1
    with pytest.raises(AttributeError):
        model.config[a2b].something = 1

    with pytest.raises(ConfigError, match="'str' is not set up for configuration"):
        print(model.config["a"].something)
    with pytest.raises(ConfigError, match="'NoneType' is not set up for configuration"):
        print(model.config[None].something)


def test_network_nesting():
    """Make sure nested networks inherit configs."""
    with nengo.Network() as net1:
        # We'll change radius and seed. Make sure we have the right defaults.
        assert nengo.Ensemble.radius.default == 1.0
        assert nengo.Ensemble.seed.default is None

        # Before = default; after = what we set
        assert net1.config[nengo.Ensemble].radius == 1.0
        assert net1.config[nengo.Ensemble].seed is None
        net1.config[nengo.Ensemble].radius = 3.0
        net1.config[nengo.Ensemble].seed = 10
        assert net1.config[nengo.Ensemble].radius == 3.0
        assert net1.config[nengo.Ensemble].seed == 10

        # If we make an ensemble, it uses what we set, of the config default
        ens1 = nengo.Ensemble(10, 1, radius=2.0)
        assert ens1.seed == 10
        assert ens1.radius == 2.0

        # It's an error to configure an actual param with the config
        with pytest.raises(ConfigError):
            net1.config[ens1].radius = 3.0

        with nengo.Network() as net2:
            # We'll just change radius in net2.

            # Before = default; after = what we set
            assert net2.config[nengo.Ensemble].radius == 1.0
            net2.config[nengo.Ensemble].radius = 5.0
            assert net2.config[nengo.Ensemble].radius == 5.0

            # If we make an ensemble, it traverses the context stack
            ens2 = nengo.Ensemble(10, 1)
            assert ens2.radius == 5.0
            assert ens2.seed == 10

            with nengo.Network() as net3:
                # Works for > 1 levels
                net3.config[nengo.Ensemble].seed = 20
                ens3 = nengo.Ensemble(10, 1)
                assert ens3.seed == 20
                assert ens3.radius == 5.0


def test_context_is_threadsafe():
    class CheckIndependence(ThreadedAssertion):
        def init_thread(self, worker):
            setattr(worker, "model", nengo.Network())
            worker.model.__enter__()

        def assert_thread(self, worker):
            assert list(nengo.Config.context) == [worker.model.config]

        def finish_thread(self, worker):
            worker.model.__exit__(*worker.exc_info)

    CheckIndependence(n_threads=2).run()


def test_defaults():
    """Test that settings defaults propagates appropriately."""
    b = nengo.Ensemble(10, dimensions=1, radius=nengo.Default, add_to_container=False)

    assert b.radius == nengo.Ensemble.radius.default

    with nengo.Network():
        c = nengo.Ensemble(10, dimensions=1, radius=nengo.Default)
        with nengo.Network() as net2:
            net2.config[nengo.Ensemble].radius = 2.0
            a = nengo.Ensemble(50, dimensions=1, radius=nengo.Default)
            del net2.config[nengo.Ensemble].radius
            d = nengo.Ensemble(50, dimensions=1, radius=nengo.Default)

    assert c.radius == nengo.Ensemble.radius.default
    assert a.radius == 2.0
    assert d.radius == nengo.Ensemble.radius.default


def test_configstack():
    """Test that setting defaults with bare configs works."""
    with nengo.Network() as net:
        net.config[nengo.Connection].transform = -1
        e1 = nengo.Ensemble(5, dimensions=1)
        e2 = nengo.Ensemble(6, dimensions=1)
        excite = nengo.Connection(e1, e2)
        with nengo.Config(nengo.Connection) as inhib:
            inhib[nengo.Connection].synapse = nengo.synapses.Lowpass(0.00848)
            inhibit = nengo.Connection(e1, e2)
    assert excite.synapse == nengo.Connection.synapse.default
    assert excite.transform.init == -1
    assert inhibit.synapse == inhib[nengo.Connection].synapse
    assert inhibit.transform.init == -1


def test_config_property():
    """Test that config can't be easily modified."""
    with nengo.Network() as net:
        with pytest.raises(ReadonlyError):
            net.config = nengo.config.Config()
        with pytest.raises(AttributeError):
            del net.config
        assert nengo.config.Config.context[-1] is net.config
    assert len(nengo.config.Config.context) == 0


def test_external_class():
    class A:
        thing = Parameter("thing", default="hey")

    inst = A()
    config = nengo.Config(A)
    config[A].set_param("amount", Parameter("amount", default=1))

    # Extra param
    assert config[inst].amount == 1

    # Default still works like Nengo object
    assert inst.thing == "hey"
    with pytest.raises(ConfigError):
        print(config[inst].thing)


def test_instance_fallthrough():
    """If the class default is set, instances should use that."""

    class A:
        pass

    inst1 = A()
    inst2 = A()
    config = nengo.Config(A)
    config[A].set_param("amount", Parameter("amount", default=1))
    assert config[A].amount == 1
    assert config[inst1].amount == 1
    assert config[inst2].amount == 1
    # Value can change for instance
    config[inst1].amount = 2
    assert config[A].amount == 1
    assert config[inst1].amount == 2
    assert config[inst2].amount == 1
    # If value to A is changed, unset instances should also change
    config[A].amount = 3
    assert config[A].amount == 3
    assert config[inst1].amount == 2
    assert config[inst2].amount == 3
    # If class default is deleted, unset instances go back
    del config[A].amount
    assert config[A].amount == 1
    assert config[inst1].amount == 2
    assert config[inst2].amount == 1


def test_contains():
    class A:
        pass

    cfg = nengo.Config(A)
    with pytest.raises(TypeError, match="Cannot check if .* is in a config"):
        print(A in cfg)

    net = nengo.Network()

    net.config[nengo.Ensemble].set_param("test", Parameter("test", None))
    assert "test" not in net.config[nengo.Ensemble]

    net.config[nengo.Ensemble].test = "testval"
    assert "test" in net.config[nengo.Ensemble]


def test_subclass_config():
    class MyParent(SupportDefaultsMixin):
        p = Parameter("p", default="baba")

        def __init__(self, p=Default):
            self.p = p

    class MyChild(MyParent):
        pass

    with nengo.Config(MyParent) as cfg:
        cfg[MyParent].p = "value1"
        a = MyChild()
        assert a.p == "value1"

    with nengo.Config(MyParent) as cfg:
        cfg[MyChild].p = "value2"
        a = MyChild()
        assert a.p == "value2"

    # If any config entry in the current context fits with the object being
    # instantiated, we use that entry, even if there's an entry that's a
    # "better" fit (i.e. same class) in a higher context.
    with nengo.Config(MyParent) as cfg1:
        cfg1[MyChild].p = "value1"

        with nengo.Config(MyParent) as cfg2:
            cfg2[MyParent].p = "value2"
            a = MyChild()
            assert a.p == "value2"


def test_classparams_del():
    """tests ClassParams.__delattr__"""
    net = nengo.Network()
    clsparams = net.config[nengo.Ensemble]

    # test that normal instance param can be added/deleted
    clsparams.set_param("test", Parameter("test", None))
    clsparams.test = "val"
    assert "test" in clsparams
    del clsparams.test
    assert "test" not in clsparams

    # test that we can set/get/delete underscore attributes regularly
    clsparams._test = 3
    assert hasattr(clsparams, "_test") and clsparams._test == 3
    del clsparams._test
    assert not hasattr(clsparams, "_test")


def test_classparams_str_repr():
    """tests the repr function in classparams class"""
    clsparams = nengo.Network().config[nengo.Ensemble]
    clsparams.set_param("test", Parameter("test", None))
    assert repr(clsparams) == "<ClassParams[Ensemble]{test: None}>"
    assert str(clsparams) == "No parameters configured for Ensemble."

    clsparams.test = "val"
    assert str(clsparams) == "Parameters configured for Ensemble:\n  test: val"


def test_config_repr():
    """tests the repr function in Config class"""
    model = nengo.Network()
    assert fnmatch(
        repr(model.config),
        "<Config at 0x* configuring (Connection, Ensemble, Node, Probe)>",
    )


def test_config_exit_errors():
    """Tests ConfigErrors in `Config.__exit__`"""
    model = nengo.Network()
    with pytest.raises(
        ConfigError, match="Config.context in bad state; was empty when exiting"
    ):
        model.config.__exit__(0, 0, 0)

    model2 = nengo.Network()
    with pytest.raises(
        ConfigError, match="Config.context in bad state; was expecting current"
    ):
        model2.config.__enter__()
        model.config.__exit__(0, 0, 0)


def test_instanceparams_str_repr():
    """Test the str and repr functions for InstanceParams class"""
    model = nengo.Network()
    with model:
        a = nengo.Ensemble(50, dimensions=1, label="a")
        model.config[nengo.Ensemble].set_param("prm", nengo.params.Parameter("prm"))
        model.config[a].prm = "val"

    assert str(model.config[a]) == "Parameters set for <Ensemble 'a'>:\n  prm: val"
    assert repr(model.config[a]) == "<InstanceParams[<Ensemble 'a'>]{prm: val}>"


def test_instanceparams_contains():
    """tests the contains function in InstanceParams class"""
    model = nengo.Network()
    model.config[nengo.Ensemble].set_param("test", Parameter("test", None))
    with model:
        a = nengo.Ensemble(5, 1)
        b = nengo.Ensemble(5, 1)
        model.config[b].test = 3

    assert "test" not in model.config[a]
    assert "test" in model.config[b]


def test_instanceparams_del():
    """tests built in params on the delattr function in InstanceParams class"""
    model = nengo.Network()
    with model:
        a = nengo.Ensemble(5, dimensions=1)

    # test that normal instance param can be added/deleted
    model.config[nengo.Ensemble].set_param("test", Parameter("test", None))
    model.config[a].test = "val"
    assert "test" in model.config[a]
    del model.config[a].test
    assert "test" not in model.config[a]

    # test that built-in parameter cannot be deleted
    with pytest.raises(ConfigError, match="Cannot configure the built-in parameter"):
        del model.config[a].bias

    # test that we can set/get/delete underscore attributes regularly
    model.config[a]._test = 3
    assert hasattr(model.config[a], "_test") and model.config[a]._test == 3
    del model.config[a]._test
    assert not hasattr(model.config[a], "_test")


def test_unconfigurable_configerror():
    """Tests exception when using settattr on something that is not configurable"""
    model = nengo.Network()
    model.config[nengo.Ensemble].set_param("prm", Parameter("prm", Unconfigurable))

    with pytest.raises(ConfigError, match="'prm' is not configurable"):
        model.config[nengo.Ensemble].prm = "other"


def test_reuse_parameters_configerror(request):
    """test that exception is raised when
    reusing parameters"""

    def finalizer():
        del nengo.Ensemble.same

    request.addfinalizer(finalizer)

    model = nengo.Network()
    nengo.Ensemble.same = Parameter("param_a")
    with pytest.raises(ConfigError, match="'same' is already a parameter in"):
        model.config[nengo.Ensemble].set_param("same", Parameter("param_b"))


def test_no_configures_args_error():
    with pytest.raises(TypeError, match="configures.* takes 1 or more arguments"):
        nengo.Network().config.configures()


def test_unconfigurable_default_configerror(request):
    """test exception when using `Config.default` with a unconfigurable parameter"""

    def finalizer():
        del nengo.Ensemble.something2

    request.addfinalizer(finalizer)

    model = nengo.Network()
    nengo.Ensemble.something2 = Parameter("something2", Unconfigurable)

    with pytest.raises(ConfigError, match="Unconfigurable parameters have no defaults"):
        model.config.default(nengo.Ensemble, "something2")


def test_get_param_on_instance_configerror():
    """test that exception is raised when getting params on instance"""
    model = nengo.Network()
    model.config[nengo.Ensemble].set_param("something", Parameter("something", None))
    with model:
        a = nengo.Ensemble(50, dimensions=1)

    with pytest.raises(ConfigError, match="Cannot get parameters on an instance"):
        model.config[a].get_param("something")


def test_config_pickle(tmp_path):
    # --- test ClassParams
    clsconfig = ClassParams(nengo.Ensemble)
    clsconfig.max_rates = nengo.dists.Choice([0])

    picklefile = str(tmp_path / "config0.pkl")
    with open(picklefile, "wb") as fh:
        pickle.dump(clsconfig, fh)

    del fh

    with open(picklefile, "rb") as fh:
        clsconfig2 = pickle.load(fh)

    assert clsconfig2.max_rates == nengo.dists.Choice([0])

    # --- test InstanceParams
    clsconfig.set_param("dummy", Parameter("dummy", default=None, optional=True))

    ens = nengo.Ensemble(10, 1, add_to_container=False)
    insconfig = InstanceParams(ens, clsconfig)
    insconfig.dummy = nengo.dists.Choice([3])

    picklefile = str(tmp_path / "config1.pkl")
    with open(picklefile, "wb") as fh:
        pickle.dump(insconfig, fh)

    del fh

    with open(picklefile, "rb") as fh:
        insconfig2 = pickle.load(fh)

    assert insconfig2.dummy == nengo.dists.Choice([3])


@pytest.mark.parametrize("simplified", [False, True])
def test_supportdefaultsmixin_exceptions(request, simplified):
    def finalizer(val=nengo.rc["exceptions"]["simplified"]):
        nengo.rc["exceptions"]["simplified"] = val

    request.addfinalizer(finalizer)

    nengo.rc["exceptions"]["simplified"] = str(simplified)

    class MyClass(SupportDefaultsMixin):
        p = Parameter("p", default="baba", readonly=True)

        def __init__(self, p=Default):
            self.p = p

    obj = MyClass()
    with pytest.raises(ReadonlyError, match="p is read-only and cannot be changed"):
        obj.p = "newval"



import threading
from collections.abc import Sequence


class ThreadLocalStack(threading.local, Sequence):  # pylint: disable=too-many-ancestors
    def __init__(self, maxsize=None):
        super().__init__()
        self.maxsize = maxsize
        self._context = []

    def __len__(self):
        return len(self._context)

    def __getitem__(self, i):
        return self._context[i]

    def append(self, item):
        if self.maxsize is not None and len(self) >= self.maxsize:
            raise RuntimeError("Stack limit exceeded.")
        self._context.append(item)

    def pop(self):
        return self._context.pop()

    def clear(self):
        self._context[:] = []



import nengo
from nengo.processes import Piecewise
from nengo.utils.numpy import rms


def test_oscillator(Simulator, plt, seed):
    model = nengo.Network(seed=seed)
    with model:
        inputs = {0: [1, 0], 0.5: [0, 0]}
        input = nengo.Node(Piecewise(inputs), label="Input")

        tau = 0.1
        freq = 5
        T = nengo.networks.Oscillator(tau, freq, n_neurons=100)
        nengo.Connection(input, T.input)

        A = nengo.Ensemble(100, dimensions=2)
        nengo.Connection(
            A, A, synapse=tau, transform=[[1, -freq * tau], [freq * tau, 1]]
        )
        nengo.Connection(input, A)

        in_probe = nengo.Probe(input, sample_every=0.01)
        A_probe = nengo.Probe(A, synapse=0.01, sample_every=0.01)
        T_probe = nengo.Probe(T.ensemble, synapse=0.01, sample_every=0.01)

    with Simulator(model) as sim:
        sim.run(3.0)

    t = sim.trange(sample_every=0.01)
    plt.plot(t, sim.data[A_probe], label="Manual")
    plt.plot(t, sim.data[T_probe], label="Template")
    plt.plot(t, sim.data[in_probe], "k", label="Input")
    plt.legend(loc="best")

    assert rms(sim.data[A_probe] - sim.data[T_probe]) < 0.2



r"""Certain features of Nengo can be configured globally through RC settings.

RC settings can be manipulated either through the ``nengo.rc`` object,
or through RC configuration files.

The ``nengo.rc`` object
=======================

The ``nengo.rc`` object gives programmatic access to
globally configured features of Nengo.

.. autodata:: nengo.rc

Configuration files
===================

``nengo.rc`` is initialized with configuration settings read
from the following files with precedence to those listed first:

1. ``nengorc`` in the current directory. This is intended to allow for project
   specific settings without hard coding them in the model script.

2. An operating system specific file in the user's home directory.

   * Windows: ``%userprofile%\.nengo\nengorc``

   * Other (OS X, Linux): ``~/.config/nengo/nengorc``

3. ``INSTALL/nengo-data/nengorc`` (where ``INSTALL`` is the
   installation directory of the Nengo package).

The RC file is divided into sections by lines containing the section name
in brackets, i.e. ``[section]``. A setting is set by giving the name followed
by a ``:`` or ``=`` and the value. All lines starting with ``#`` or ``;`` are
comments.

For example, to set the size of the decoder cache to 512 MB,
add the following to a configuration file:

.. code-block:: ini

   [decoder_cache]
   size = 512 MB

Configuration options
=====================

All of the configuration options are listed in the example
configuration file, which is included with Nengo
and copied below.

Commented lines show the default values for each setting.

.. _nengorc:

.. include:: ../nengo-data/nengorc
   :literal:
   :start-line: 29

"""

import logging
from configparser import DEFAULTSECT, ConfigParser

import numpy as np

import nengo.utils.paths

logger = logging.getLogger(__name__)

# The default core Nengo RC settings. Access with
#   nengo.RC_DEFAULTS[section_name][option_name]
RC_DEFAULTS = {
    "precision": {"bits": 64},
    "decoder_cache": {
        "enabled": True,
        "readonly": False,
        "size": "512 MB",
        "path": nengo.utils.paths.decoder_cache_dir,
    },
    "progress": {"progress_bar": "auto"},
    "exceptions": {"simplified": True},
    "nengo.Simulator": {"fail_fast": False},
}

# The RC files in the order in which they will be read.
RC_FILES = [
    nengo.utils.paths.nengorc["system"],
    nengo.utils.paths.nengorc["user"],
    nengo.utils.paths.nengorc["project"],
]


class _RC(ConfigParser):  # pylint: disable=too-many-ancestors
    """Allows reading from and writing to Nengo RC settings.

    This object is a :class:`configparser.ConfigParser`, which means that
    values can be accessed and manipulated like a dictionary:

    .. testcode::

       oldsize = nengo.rc["decoder_cache"]["size"]
       nengo.rc["decoder_cache"]["size"] = "2 GB"

    All values are stored as strings. If you want to store or retrieve a
    specific datatype, you should coerce it appropriately (e.g., with ``int()``).
    Booleans are more flexible, so you should use the ``getboolean`` method
    to access boolean values.

    .. testcode::

       simple = nengo.rc["exceptions"].getboolean("simplified")

    In addition to the normal :class:`configparser.ConfigParser` methods,
    this object also has a ``reload_rc`` method to reset ``nengo.rc``
    to default settings:

    .. testcode::

       nengo.rc.reload_rc()  # Reads defaults from configuration files
       nengo.rc.reload_rc(filenames=[])  # Ignores configuration files

    """

    def __init__(self):
        super().__init__()
        self.reload_rc()

    @property
    def float_dtype(self):
        bits = self.get("precision", "bits")
        return np.dtype(f"float{bits}")

    @property
    def int_dtype(self):
        bits = self.get("precision", "bits")
        return np.dtype(f"int{bits}")

    def _clear(self):
        self.remove_section(DEFAULTSECT)
        for s in self.sections():
            self.remove_section(s)

    def _init_defaults(self):
        for section, settings in RC_DEFAULTS.items():
            self.add_section(section)
            for k, v in settings.items():
                self.set(section, k, str(v))

    def read_file(self, fp, filename=None):  # pylint: disable=arguments-renamed
        if filename is None:
            filename = fp.name if hasattr(fp, "name") else "<???>"

        logger.debug("Reading configuration from %s", filename)
        return super().read_file(fp, filename)

    def read(self, filenames):
        logger.debug("Reading configuration files %s", filenames)
        return super().read(filenames)

    def reload_rc(self, filenames=None):
        """Resets the currently loaded RC settings and loads new RC files.

        Parameters
        ----------
        filenames: iterable object
            Filenames of RC files to load.
        """
        if filenames is None:
            filenames = RC_FILES

        self._clear()
        self._init_defaults()
        self.read(filenames)


rc = _RC()



import numpy as np
import pytest

from nengo.utils.testing import ThreadedAssertion, signals_allclose


def add_close_noise(x, atol, rtol, rng):
    scale = rng.uniform(-rtol, rtol, size=x.shape)
    offset = rng.uniform(-atol, atol, size=x.shape)
    return x + scale * np.abs(x) + offset


@pytest.mark.parametrize("multiple_targets", (False, True))
def test_signals_allclose(multiple_targets, rng):
    nt = 100  # signal length
    ny = 3  # number of comparison signals
    atol = 1e-5
    rtol = 1e-3

    t = 0.01 * np.arange(nt)
    x = rng.uniform(-1, 1, size=(nt, ny if multiple_targets else 1))
    get_y = lambda close: add_close_noise(
        x, (1 if close else 3) * atol, (1 if close else 3) * rtol, rng
    )
    y_close = (
        get_y(True)
        if multiple_targets
        else np.column_stack([get_y(True) for _ in range(ny)])
    )
    y_far = (
        get_y(False)
        if multiple_targets
        else np.column_stack([get_y(False) for _ in range(ny)])
    )

    result = signals_allclose(
        t, x, y_close, atol=atol, rtol=rtol, individual_results=False
    )
    assert result is True

    result = signals_allclose(
        t, x, y_far, atol=atol, rtol=rtol, individual_results=False
    )
    assert result is False

    result = signals_allclose(
        t, x, y_close, atol=atol, rtol=rtol, individual_results=True
    )
    assert np.array_equal(result, np.ones(ny, dtype=bool))

    result = signals_allclose(
        t, x, y_far, atol=atol, rtol=rtol, individual_results=True
    )
    assert np.array_equal(result, np.zeros(ny, dtype=bool))

    # tests for `signals.ndim == 1` and `targets.ndim == 1`
    if not multiple_targets:
        assert x.shape[1] == 1
        assert signals_allclose(t, x[:, 0], y_close[:, 0], atol=atol, rtol=rtol)
        assert not signals_allclose(t, x[:, 0], y_far[:, 0], atol=atol, rtol=rtol)


@pytest.mark.parametrize("ny", (1, 3))
def test_signals_allclose_plot(ny, rng, plt):
    if plt.__file__ == "/dev/null":
        pytest.skip("Only runs when plotting is enabled")

    nt = 100  # signal length
    atol = 1e-5
    rtol = 1e-3

    t = 0.01 * np.arange(nt)
    x = rng.uniform(-1, 1, size=(nt, ny))
    y = add_close_noise(x, atol, rtol, rng)
    labels = [f"lab{i}" for i in range(ny)]

    if ny == 1:
        x, y, labels = x[:, 0], y[:, 0], labels[0]

    fig = plt.figure()
    result = signals_allclose(t, x, y, atol=atol, rtol=rtol, plt=plt, labels=labels)

    # check the legend
    legend = fig.axes[0].get_legend()
    for i in range(ny):
        ref_text = labels if ny == 1 else labels[i]
        assert legend.texts[i].get_text() == ref_text

    assert result


def test_threadedassertion():
    class Test(ThreadedAssertion):
        def __init__(self, n_threads, assert_inds, **kwargs):
            self.assert_inds = assert_inds
            super().__init__(n_threads, **kwargs)

        def init_thread(self, worker):
            worker.init_param = "good"

        def assert_thread(self, worker):
            assert worker.init_param == "good", "Worker did not init"

            if worker.n in self.assert_inds:
                raise RuntimeError(f"Worker {worker.n} failed properly")

        def finish_thread(self, worker):
            worker.finish_param = "finished"

    # running with assert_inds=[] should pass
    test = Test(n_threads=3, assert_inds=[])
    test.run()
    assert all(thread.finish_param == "finished" for thread in test.threads)

    with pytest.raises(RuntimeError, match="Worker 1 failed properly"):
        Test(n_threads=3, assert_inds=[1]).run()

    test = Test(n_threads=3, assert_inds=[1])
    with pytest.raises(RuntimeError, match="Worker 1 failed properly"):
        test.run()
    assert all(thread.finish_param == "finished" for thread in test.threads)

    test = Test(n_threads=3, assert_inds=[2])
    with pytest.raises(RuntimeError, match="Worker 2 failed properly"):
        test.run()
    assert all(thread.finish_param == "finished" for thread in test.threads)



import numpy as np
import pytest

import nengo
from nengo.builder.neurons import SimNeurons
from nengo.builder.tests.test_operator import _test_operator_arg_attributes


@pytest.mark.parametrize(
    "SpikingType",
    (nengo.RegularSpiking, nengo.PoissonSpiking, nengo.StochasticSpiking),
)
def test_spiking_builders(SpikingType):
    # use a base type with its own state(s), to make sure those states get built
    neuron_type = SpikingType(
        nengo.AdaptiveLIFRate(initial_state={"adaptation": np.ones(10) * 2}),
        initial_state={"voltage": np.ones(10) * 3}
        if SpikingType is nengo.RegularSpiking
        else {},
    )

    with nengo.Network() as net:
        neurons = nengo.Ensemble(10, 1, neuron_type=neuron_type).neurons

        # check that the expected attributes are probeable
        nengo.Probe(neurons, "output")
        nengo.Probe(neurons, "rate_out")
        nengo.Probe(neurons, "adaptation")
        if SpikingType is nengo.RegularSpiking:
            nengo.Probe(neurons, "voltage")

    with nengo.Simulator(net) as sim:
        ops = [op for op in sim.step_order if isinstance(op, SimNeurons)]
        assert len(ops) == 2
        assert ops[0].neurons is neurons.ensemble.neuron_type.base_type
        assert ops[1].neurons is neurons.ensemble.neuron_type

        adaptation = sim.model.sig[neurons]["adaptation"]
        # All signals get put in `sets`
        assert sum(adaptation is sig for sig in ops[0].sets) == 1

        # check that initial state argument is applied correctly
        assert np.allclose(sim.signals[sim.model.sig[neurons]["adaptation"]], 2)
        if SpikingType is nengo.RegularSpiking:
            assert np.allclose(sim.signals[sim.model.sig[neurons]["voltage"]], 3)


def test_simneurons():
    argnames = ["neurons", "J", "output"]
    non_signals = ["neurons"]
    _, sim = _test_operator_arg_attributes(
        SimNeurons, argnames, non_signals=non_signals
    )

    assert str(sim) == "SimNeurons{neurons, J, output}"



"""A pytest plugin to support backends running the Nengo test suite.

Unsupported tests
-----------------
The ``nengo_test_unsupported`` option allows you to specify Nengo tests
unsupported by a particular simulator.
This is used if you are writing a backend and want to ignore
tests for functions that your backend currently does not support.

Each line represents one test pattern to skip,
and must be followed by a line containing a string in quotes
denoting the reason for skipping the test(s).

The pattern uses
`Unix filename pattern matching
<https://docs.python.org/3/library/fnmatch.html>`_,
including wildcard characters ``?`` and ``*`` to match one or more characters.
The pattern matches to the test name,
which is the same as the pytest ``nodeid``
seen when calling pytest with the ``-v`` argument.

.. code-block:: ini

   nengo_test_unsupported =
       nengo/tests/test_file_path.py::test_function_name
           "This is a message giving the reason we skip this test"
       nengo/tests/test_file_two.py::test_other_thing
           "This is a test with a multi-line reason for skipping.
           Make sure to use quotes around the whole string (and not inside)."
       nengo/tests/test_file_two.py::test_parametrized_thing[param_value]
           "This skips a parametrized test with a specific parameter value."
"""

import importlib
import shlex
import sys
from fnmatch import fnmatch

try:
    import resource
except ImportError:  # pragma: no cover
    resource = None  # `resource` not available on Windows

import pytest


def is_sim_overridden(config):
    return config.getini("nengo_simulator") != "nengo.Simulator" or config.getini(
        "nengo_simloader"
    )


def is_nengo_test(item):
    return str(item.fspath.pypkgpath()).endswith("nengo")


def deselect_by_condition(condition, items, config):
    remaining = []
    deselected = []
    for item in items:
        if is_nengo_test(item) and condition(item):
            deselected.append(item)
        else:
            remaining.append(item)

    if deselected:
        config.hook.pytest_deselected(items=deselected)
        items[:] = remaining


def load_class(fully_qualified_name):
    mod_name, cls_name = fully_qualified_name.rsplit(".", 1)
    mod = importlib.import_module(mod_name)
    return getattr(mod, cls_name)


def pytest_configure(config):
    if config.getoption("memory") and resource is None:  # pragma: no cover
        raise ValueError("'--memory' option not supported on this platform")

    config.addinivalue_line("markers", "example: Mark a test as an example.")
    config.addinivalue_line(
        "markers", "slow: Mark a test as slow to skip it per default."
    )


def pytest_addoption(parser):
    parser.addoption(
        "--unsupported",
        action="store_true",
        default=False,
        help="Run (with xfail) tests marked as unsupported by this backend.",
    )
    parser.addoption(
        "--noexamples", action="store_true", default=False, help="Do not run examples"
    )
    parser.addoption(
        "--slow", action="store_true", default=False, help="Also run slow tests."
    )
    parser.addoption(
        "--spa", action="store_true", default=False, help="Run deprecated SPA tests"
    )
    group = parser.getgroup("terminal reporting", "reporting", after="general")
    group.addoption(
        "--memory",
        action="store_true",
        default=False,
        help="Show memory consumed by Python after all tests are run "
        "(not available on Windows)",
    )

    parser.addini(
        "nengo_simulator", default="nengo.Simulator", help="The simulator class to test"
    )
    parser.addini(
        "nengo_simloader",
        default=None,
        help="A function that returns the simulator class to test",
    )
    parser.addini(
        "nengo_neurons",
        type="linelist",
        default=[
            "nengo.Direct",
            "nengo.LIF",
            "nengo.LIFRate",
            "nengo.RectifiedLinear",
            "nengo.Sigmoid",
            "nengo.SpikingRectifiedLinear",
            "nengo.Tanh",
            "nengo.tests.test_neurons.SpikingTanh",
        ],
        help="Neuron types under test",
    )
    parser.addini(
        "nengo_test_unsupported",
        type="linelist",
        default=[],
        help="List of unsupported unit tests with reason for exclusion",
    )


def pytest_generate_tests(metafunc):
    marks = [
        getattr(pytest.mark, m.name)(*m.args, **m.kwargs)
        for m in getattr(metafunc.function, "pytestmark", [])
    ]

    def mark_neuron_type(NeuronType):
        if NeuronType.__name__ == "Sigmoid":
            NeuronType = pytest.param(
                NeuronType,
                marks=[pytest.mark.filterwarnings("ignore:overflow encountered in exp")]
                + marks,
            )
        return NeuronType

    all_neuron_types = [load_class(n) for n in metafunc.config.getini("nengo_neurons")]

    if "AnyNeuronType" in metafunc.fixturenames:
        metafunc.parametrize(
            "AnyNeuronType",
            [mark_neuron_type(NeuronType) for NeuronType in all_neuron_types],
        )
    if "NonDirectNeuronType" in metafunc.fixturenames:
        metafunc.parametrize(
            "NonDirectNeuronType",
            [
                mark_neuron_type(NeuronType)
                for NeuronType in all_neuron_types
                if NeuronType.__name__ != "Direct"
            ],
        )
    if "PositiveNeuronType" in metafunc.fixturenames:
        metafunc.parametrize(
            "PositiveNeuronType",
            [
                mark_neuron_type(NeuronType)
                for NeuronType in all_neuron_types
                if NeuronType.__name__ != "Direct" and not NeuronType.negative
            ],
        )


def pytest_collection_modifyitems(session, config, items):
    uses_sim = lambda item: "Simulator" in item.fixturenames
    if is_sim_overridden(config):
        deselect_by_condition(lambda item: not uses_sim(item), items, config)
    if config.getvalue("noexamples"):
        deselect_by_condition(
            lambda item: item.get_closest_marker("example"), items, config
        )
    if not config.getvalue("slow"):
        skip_slow = pytest.mark.skip("slow tests not requested")
        for item in items:
            if item.get_closest_marker("slow"):
                item.add_marker(skip_slow)
    if not config.getvalue("spa"):
        deselect_by_condition(lambda item: "spa/tests" in item.nodeid, items, config)


def pytest_report_collectionfinish(config, startdir, items):
    if not any(is_nengo_test(item) for item in items):
        return

    deselect_reasons = ["Nengo core tests collected"]
    if is_sim_overridden(config):
        deselect_reasons.append(
            " frontend tests deselected because simulator is not nengo.Simulator"
        )
    if config.getvalue("noexamples"):
        deselect_reasons.append(" example tests deselected (--noexamples passed)")
    if not config.getvalue("slow"):
        deselect_reasons.append(" slow tests skipped (pass --slow to run them)")
    if not config.getvalue("spa"):
        deselect_reasons.append(" spa tests deselected (pass --spa to run them)")

    return deselect_reasons


def pytest_runtest_setup(item):
    # join all the lines and then split (preserving quoted strings)
    unsupported = shlex.split(" ".join(item.config.getini("nengo_test_unsupported")))
    # group pairs (representing testname + reason)
    unsupported = [unsupported[i : i + 2] for i in range(0, len(unsupported), 2)]

    for test, reason in unsupported:
        # wrap square brackets to interpret them literally
        # (see https://docs.python.org/3/library/fnmatch.html)
        test = "".join(f"[{c}]" if c in ("[", "]") else c for c in test)

        # We add a '*' before test to eliminate the surprise of needing
        # a '*' before the name of a test function.
        test = "*" + test

        if is_nengo_test(item) and fnmatch(item.nodeid, test):
            if item.config.getvalue("unsupported"):
                item.add_marker(pytest.mark.xfail(reason=reason))
            else:
                pytest.skip(reason)


def pytest_terminal_summary(terminalreporter):
    if resource and terminalreporter.config.option.memory:
        # Calculate memory usage; details at
        # http://fa.bianp.net/blog/2013/different-ways-to-get-memory-consumption-or-lessons-learned-from-memory_profiler/  # noqa, pylint: disable=line-too-long
        rusage_denom = 1024.0
        if sys.platform == "darwin":  # pragma: no cover
            # ... it seems that in OSX the output is in different units ...
            rusage_denom = rusage_denom * rusage_denom
        mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / rusage_denom
        terminalreporter.write_sep("=", f"total memory consumed: {mem:.2f} MiB")

        # Ensure we only print once
        terminalreporter.config.option.memory = False


@pytest.fixture(scope="session")
def Simulator(request):
    """The Simulator class being tested.

    Please use this, and not ``nengo.Simulator`` directly.
    """

    if request.config.getini("nengo_simloader"):
        # Note: --simloader takes precedence over --simulator.
        # Some backends might specify both for backwards compatibility reasons.
        SimLoader = load_class(request.config.getini("nengo_simloader"))
        return SimLoader(request)
    else:
        return load_class(request.config.getini("nengo_simulator"))



import pathlib
import sys

if sys.platform.startswith("win"):  # pragma: no cover
    config_dir = pathlib.Path.home() / ".nengo"
    cache_dir = config_dir / "cache"
else:
    config_dir = pathlib.Path.home() / ".config" / "nengo"
    cache_dir = pathlib.Path.home() / ".cache" / "nengo"

decoder_cache_dir = cache_dir / "decoders"
install_dir = pathlib.Path(__file__).parents[2]
data_dir = install_dir / "nengo" / "data"
examples_dir = install_dir / "docs" / "examples"

nengorc = {
    "system": install_dir / "nengo-data" / "nengorc",
    "user": config_dir / "nengorc",
    "project": pathlib.Path.cwd() / "nengorc",
}



import pytest

import nengo
from nengo.utils.network import activate_direct_mode


def test_withself():
    model = nengo.Network(label="test_withself")
    with model:
        n1 = nengo.Node(0.5)
        assert n1 in model.nodes
        e1 = nengo.Ensemble(10, dimensions=1)
        assert e1 in model.ensembles
        c1 = nengo.Connection(n1, e1)
        assert c1 in model.connections
        ea1 = nengo.networks.EnsembleArray(10, n_ensembles=2)
        assert ea1 in model.networks
        assert len(ea1.ensembles) == 2
        n2 = ea1.add_output("out", None)
        assert n2 in ea1.nodes
        with ea1:
            e2 = nengo.Ensemble(10, dimensions=1)
            assert e2 in ea1.ensembles
    assert len(nengo.Network.context) == 0


def test_activate_direct_mode():
    with nengo.Network() as model:
        direct_mode_ens = [nengo.Ensemble(10, 1), nengo.Ensemble(10, 1)]
        non_direct_pre = nengo.Ensemble(10, 1)
        non_direct_post = nengo.Ensemble(10, 1)
        non_direct_probe = nengo.Ensemble(10, 1)
        non_direct_mode_ens = [non_direct_pre, non_direct_post, non_direct_probe]

        nengo.Connection(direct_mode_ens[0], direct_mode_ens[1])

        nengo.Connection(non_direct_pre.neurons[0], direct_mode_ens[0])
        nengo.Connection(direct_mode_ens[1], non_direct_post.neurons[0])
        nengo.Probe(non_direct_probe.neurons)

    activate_direct_mode(model)

    for ens in direct_mode_ens:
        assert type(ens.neuron_type) is nengo.Direct
    for ens in non_direct_mode_ens:
        assert type(ens.neuron_type) is not nengo.Direct


@pytest.mark.parametrize(
    "learning_rule, weights",
    (
        (nengo.PES(), False),
        (nengo.BCM(), True),
        (nengo.Oja(), True),
        (nengo.Voja(), False),
    ),
)
def test_activate_direct_mode_learning(Simulator, learning_rule, weights):
    with nengo.Network() as model:
        pre = nengo.Ensemble(10, 1)
        post = nengo.Ensemble(10, 1)
        conn = nengo.Connection(
            pre, post, solver=nengo.solvers.LstsqL2(weights=weights)
        )
        conn.learning_rule_type = learning_rule

    activate_direct_mode(model)

    with Simulator(model) as sim:
        sim.run(0.01)



import numpy as np
import pytest

import nengo
from nengo.builder import Model
from nengo.builder.connection import build_linear_system
from nengo.exceptions import BuildError


def test_build_linear_system(seed, rng, plt, allclose):
    func = lambda x: x ** 2

    with nengo.Network(seed=seed) as net:
        conn = nengo.Connection(
            nengo.Ensemble(60, 1), nengo.Ensemble(50, 1), function=func
        )

    model = Model()
    model.build(net)
    eval_points, activities, targets = build_linear_system(model, conn, rng)
    assert eval_points.shape[1] == 1
    assert targets.shape[1] == 1
    assert activities.shape[1] == 60
    assert eval_points.shape[0] == activities.shape[0] == targets.shape[0]

    X = eval_points
    AA = activities.T.dot(activities)
    AX = activities.T.dot(eval_points)
    AY = activities.T.dot(targets)
    WX = np.linalg.solve(AA, AX)
    WY = np.linalg.solve(AA, AY)

    Xhat = activities.dot(WX)
    Yhat = activities.dot(WY)

    i = np.argsort(eval_points.ravel())
    plt.plot(X[i], Xhat[i])
    plt.plot(X[i], Yhat[i])

    assert allclose(Xhat, X, atol=1e-1)
    assert allclose(Yhat, func(X), atol=1e-1)


def test_build_linear_system_zeroact(seed, rng):
    eval_points = np.linspace(-0.1, 0.1, 100)[:, None]

    with nengo.Network(seed=seed) as net:
        a = nengo.Ensemble(5, 1, intercepts=nengo.dists.Choice([0.9]))
        b = nengo.Ensemble(5, 1, intercepts=nengo.dists.Choice([0.9]))

    model = Model()
    model.build(net)

    conn = nengo.Connection(a, b, eval_points=eval_points, add_to_container=False)
    with pytest.raises(BuildError, match="'activities' matrix is all zero"):
        build_linear_system(model, conn, rng)


def test_build_connection_errors():
    # --- test function points on direct connection error
    with nengo.Network() as net:
        a = nengo.Ensemble(5, 1, neuron_type=nengo.Direct())
        b = nengo.Ensemble(4, 1)
        nengo.Connection(a, b, eval_points=np.ones((3, 1)), function=np.ones((3, 1)))

    with pytest.raises(BuildError, match="Cannot use function points.*direct conn"):
        with nengo.Simulator(net):
            pass

    # --- test connection to object not in the model
    b = nengo.Ensemble(5, 1, add_to_container=False)
    with nengo.Network() as net:
        a = nengo.Ensemble(5, 1)
        conn = nengo.Connection(a, b)

    with pytest.raises(BuildError, match="is not in the model"):
        with nengo.Simulator(net):
            pass

    # --- test connection with post object of size zero
    with nengo.Network() as net:
        a = nengo.Ensemble(5, 1)
        b = nengo.Node([0])
        conn = nengo.Connection(a, a)

    # hack to get around API validation
    nengo.Connection.post.data[conn] = b

    with pytest.raises(BuildError, match="the 'post' object.*has a 'in' size of zero"):
        with nengo.Simulator(net):
            pass



#!/usr/bin/env python

# Automatically generated by nengo-bones, do not edit this file directly

import io
import pathlib
import runpy

try:
    from setuptools import find_packages, setup
except ImportError:
    raise ImportError(
        "'setuptools' is required but not installed. To install it, "
        "follow the instructions at "
        "https://pip.pypa.io/en/stable/installing/#installing-with-get-pip-py"
    )


def read(*filenames, **kwargs):
    encoding = kwargs.get("encoding", "utf-8")
    sep = kwargs.get("sep", "\n")
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


root = pathlib.Path(__file__).parent
version = runpy.run_path(str(root / "nengo" / "version.py"))["version"]

install_req = [
    "numpy>=1.19",
]
docs_req = [
    "sphinx>=1.8",
    "jupyter",
    "matplotlib>=1.4",
    "nbsphinx",
    "numpydoc>=0.6",
    "nengo_sphinx_theme>=1.2.0",
]
optional_req = [
    "scipy>=0.13",
    "scikit-learn",
]
tests_req = [
    "jupyter",
    "matplotlib>=1.4",
    "pytest>=3.9.1",
    "pytest-allclose",
    "pytest-plt",
    "pytest-rng",
]

setup(
    name="nengo",
    version=version,
    author="Applied Brain Research",
    author_email="info@appliedbrainresearch.com",
    packages=find_packages(),
    url="https://www.nengo.ai/nengo",
    include_package_data=True,
    license="Free for non-commercial use",
    description="Tools for building and simulating large-scale neural models",
    long_description=read("README.rst", "CHANGES.rst"),
    zip_safe=False,
    install_requires=install_req,
    extras_require={
        "all": docs_req + optional_req + tests_req,
        "docs": docs_req,
        "optional": optional_req,
        "tests": tests_req,
    },
    python_requires=">=3.6",
    package_data={
        "nengo": [
            "nengo-data/nengorc",
        ],
    },
    py_modules=[
        "pytest_nengo",
    ],
    entry_points={
        "nengo.backends": [
            "reference = nengo:Simulator",
        ],
        "pytest11": [
            "nengo = pytest_nengo",
        ],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Nengo",
        "Intended Audience :: Science/Research",
        "License :: Free for non-commercial use",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)
